from octopus.core.instruction import Instruction


class EvmInstruction(Instruction):
    """ETH Instruction

    TODO

    """

    def __init__(self, opcode, name,
                 operand_size, pops, pushes, fee,
                 description, operand=None,
                 operand_interpretation=None, offset=0, xref=None):
        """ TODO """
        super().__init__(opcode=opcode, name=name,
                         operand_size=operand_size, pops=pops, pushes=pushes,
                         fee=fee, description=description, operand=operand,
                         operand_interpretation=operand_interpretation,
                         offset=offset, xref=xref)

    @property
    def group(self):
        '''Instruction classification as per the yellow paper'''
        classes = {0: 'Stop and Arithmetic Operations',
                   1: 'Comparison & Bitwise Logic Operations',
                   2: 'SHA3',
                   3: 'Environmental Information',
                   4: 'Block Information',
                   5: 'Stack, Memory, Storage and Flow Operations',
                   6: 'Push Operations',
                   7: 'Push Operations',
                   8: 'Duplication Operations',
                   9: 'Exchange Operations',
                   0xa: 'Logging Operations',
                   0xf: 'System operations'}
        return classes.get(self.opcode >> 4, 'Invalid instruction')

    @property
    def is_terminator(self):
        """ True if the instruction is a basic block terminator """
        return self.is_branch or self.is_halt

    @property
    def is_branch_conditional(self):
        """ Return list if the instruction is a jump """
        return self.semantics in ['JUMPI']

    @property
    def is_branch_unconditional(self):
        """ Return list if the instruction is a jump """
        return self.semantics in ['JUMP']

    @property
    def is_system(self):
        """ True if the instruction is a system operation """
        return self.group == 'System operations'

    @property
    def is_arithmetic(self):
        """ True if the instruction is an arithmetic operation """
        return self.group == 'Stop and Arithmetic Operations'

    @property
    def is_comparaison_logic(self):
        """ True if the instruction is a Comparison & Bitwise Logic Operations """
        return self.group == 'Comparison & Bitwise Logic Operations'

    @property
    def is_sha3(self):
        """ True if the instruction is SHA3"""
        return self.group == 'SHA3'

    @property
    def is_environmental(self):
        """ True if the instruction access enviromental data """
        return self.group == 'Environmental Information'

    @property
    def uses_block_info(self):
        """ True if the instruction access block information """
        return self.group == 'Block Information'

    @property
    def uses_stack_block_storage_info(self):
        """ True if the instruction are in the group Stack, Memory, Storage and Flow Operations """
        return self.group == 'Stack, Memory, Storage and Flow Operations'

    @property
    def is_push(self):
        """ True if the instruction is a push Operations """
        return self.group == 'Push Operations'

    @property
    def have_xref(self):
        """ True if the instruction is a basic block terminator """
        return self.is_branch

    def set_xref(self, v):
        """ TODO """
        self._xref = int.from_bytes(v, byteorder='big')

    @property
    def is_halt(self):
        """ Return list if the instruction is a basic block terminator """
        return self.semantics in ('RETURN', 'STOP', 'INVALID',
                                  'SELFDESTRUCT', 'REVERT')




def bytecode_to_bytes(bytecode):
    if str(bytecode).startswith("0x"):
        bytecode = str(bytecode)[2:]

    try:
        bytecode = bytes.fromhex(bytecode)
    except TypeError:
        pass
    return bytecode



from octopus.core.instruction import Instruction

from wasm.opcodes import (INSN_ENTER_BLOCK,
                          INSN_LEAVE_BLOCK,
                          INSN_NO_FLOW)  # INSN_BRANCH


class WasmInstruction(Instruction):
    """Wasm Instruction

    TODO

    """
    def __init__(self, opcode, name, imm_struct, operand_size, flags, insn_byte,
                 pops, pushes, description, operand_interpretation=None, offset=0):
        """ TODO """
        self.opcode = opcode
        self.offset = offset
        self.name = name
        self.description = description
        self.operand_size = operand_size
        self.flags = flags
        self.operand = insn_byte  # Immediate operand if any
        # specific interpretation of operand value
        self.operand_interpretation = operand_interpretation
        self.insn_byte = insn_byte
        self.pops = pops
        self.pushes = pushes
        self.imm_struct = imm_struct
        self.xref = list()
        self.ssa = None

    def __eq__(self, other):
        """ Instructions are equal if all features match  """
        return self.opcode == other.opcode and\
            self.name == other.name and\
            self.insn_byte == other.insn_byte and\
            self.operand_size == other.operand_size and\
            self.pops == other.pops and\
            self.pushes == other.pushes and\
            self.operand_interpretation == other.operand_interpretation and\
            self.description == other.description

    def __str__(self):
        """ String representation of the instruction """
        if self.operand_interpretation:
            return self.operand_interpretation
        # elif self.operand:
        #    return self.name + str(self.operand)
        else:
            return self.name

    @property
    def group(self):
        """ Instruction classification per group """
        classes = {0x00: 'Control',
                   0x1A: 'Parametric',
                   0x20: 'Variable',
                   0x28: 'Memory',
                   0x41: 'Numeric'}

        last_class = classes.get(0)
        for k, v in classes.items():
            if self.opcode >= k:
                last_class = v
            else:
                return last_class

    @property
    def is_branch_conditional(self):
        """ Return True if the instruction is a conditional jump """
        return self.name in ['br_if', 'br_table', 'if']

    @property
    def is_branch_unconditional(self):
        """ Return True if the instruction is a unconditional jump """
        return self.name in ['br']

    @property
    def is_call(self):
        """ True if the instruction is a call instruction """
        return self.name in ['call', 'call_indirect']

    @property
    def is_branch(self):
        return self.is_branch_conditional or self.is_branch_unconditional

    @property
    def is_halt(self):
        """ Return True if the instruction is a branch terminator """
        return self.flags & INSN_NO_FLOW
        # return self.name in ['return']  # unreachable

    @property
    def is_terminator(self):
        """ True if the instruction is a basic block terminator """
        return self.is_branch or self.is_halt

    @property
    def is_block_starter(self):
        """ Return True if the instruction is a basic block starter """
        return self.flags & INSN_ENTER_BLOCK

    @property
    def is_block_terminator(self):
        """ Return True if the instruction is a basic block terminator """
        return self.flags & INSN_LEAVE_BLOCK



from octopus.core.utils import bytecode_to_bytes


class BytecodeEmptyException(Exception):
    """Exception raised when bytecode is None"""
    pass


class Disassembler(object):
    """ Generic Disassembler class """

    def __init__(self, asm, bytecode=None):
        self.bytecode = bytecode
        self.instructions = list()
        self.reverse_instructions = dict()
        self.asm = asm

    def attributes_reset(self):
        """Reset instructions class attributes """
        self.instructions = list()
        self.reverse_instructions = dict()

    def disassemble_opcode(self, bytecode, offset=0):
        """ Generic method to disassemble one instruction """
        raise NotImplementedError

    def disassemble(self, bytecode=None, offset=0, r_format='list'):
        """Generic method to disassemble bytecode

        :param bytecode: bytecode sequence
        :param offset: start offset
        :param r_format: output format ('list'/'text'/'reverse')
        :type bytecode: bytes, str
        :type offset: int
        :type r_format: list, str, dict
        :return: dissassembly result depending of r_format
        :rtype: list, str, dict
        """
        # reinitialize class variable
        self.attributes_reset()

        self.bytecode = bytecode if bytecode else self.bytecode
        if not self.bytecode:
            raise BytecodeEmptyException()

        self.bytecode = bytecode_to_bytes(self.bytecode)

        while offset < len(self.bytecode):
            instr = self.disassemble_opcode(self.bytecode[offset:], offset)
            offset += instr.size
            self.instructions.append(instr)

        # fill reverse instructions
        self.reverse_instructions = {k: v for k, v in
                                     enumerate(self.instructions)}

        # return instructions
        if r_format == 'list':
            return self.instructions
        elif r_format == 'text':
            return '\n'.join(map(str, self.instructions))
        elif r_format == 'reverse':
            return self.reverse_instructions

    def disassemble_contract(self, contract):
        """ Generic method to disassemble a Contract """

        # reinitialize class variable
        self.attributes_reset()
        # update class bytecode
        self.bytecode = contract.bytecode

        self.disassemble(self.bytecode)



from octopus.core.instruction import Instruction

import struct


class BitcoinInstruction(Instruction):
    """BTC Instruction

    TODO

    """
    def __init__(self, opcode, name,
                 operand_size, pops, pushes, fee,
                 description, operand=None,
                 operand_interpretation=None,
                 offset=0, xref=None):
        """ TODO """
        super().__init__(opcode=opcode, name=name,
                         operand_size=operand_size, pops=pops, pushes=pushes,
                         fee=fee, description=description, operand=operand,
                         operand_interpretation=operand_interpretation,
                         offset=offset, xref=xref)

    # redefinition of this function
    def __str__(self):
        """ String representation of the instruction """
        output = ''
        if self.has_operand:
            if type(self.operand) == int:
                output = self.name + '' + self.operand.hex()
            else:
                if self.operand_size < 5:  # in ['', 'OP_PUSHDATA1', 'OP_PUSHDATA2']:
                    output = '%d' % int.from_bytes(self.operand,
                                                   byteorder='little', signed=True)
                elif self.name == '' or self.has_length_operand:
                    output = self.operand.hex()
                else:
                    output = self.name + ' ' + self.operand.hex()
        else:
            # format to be compliant with result from RPC call
            if (self.opcode >= 0x51 and self.opcode < 0x5A) or self.name == 'OP_0':
                output = self.name[-1]
            elif self.opcode >= 0x5A and self.opcode <= 0x60:
                output = self.name[-2:]
            else:
                output = self.name
        return output

    @property
    def has_length_operand(self):
        return self.name in ['OP_PUSHDATA1', 'OP_PUSHDATA2', 'OP_PUSHDATA4']

    def format_operand_size(self):
        if self.name == 'OP_PUSHDATA1':
            self.operand_size = struct.unpack('<B', self.operand_size)[0]
        elif self.name == 'OP_PUSHDATA2':
            self.operand_size = struct.unpack(b"<H", self.operand_size)[0]
        elif self.name == 'OP_PUSHDATA4':
            self.operand_size = struct.unpack(b"<I", self.operand_size)[0]
        else:
            pass

    def format_length_operand(self, value):
        if self.name == 'OP_PUSHDATA1':
            return value.to_bytes(1, 'little')
        elif self.name == 'OP_PUSHDATA2':
            return value.to_bytes(2, 'little')
        elif self.name == 'OP_PUSHDATA4':
            return value.to_bytes(4, 'little')
        else:
            pass

    # overwrite
    @property
    def bytes(self):
        """ Encoded instruction """
        byte = bytearray()
        byte.append(self.opcode)
        if self.has_length_operand:
            [byte.append(x) for x in self.format_length_operand(len(self.operand))]
        if self.operand:
            [byte.append(x) for x in self.operand]
        return "".join(map(chr, byte))

    # overwrite
    @property
    def size(self):
        """ Size of the encoded instruction """
        if self.has_length_operand:
            return self.operand_size + self.operand_size + 1
        else:
            return self.operand_size + 1



from octopus.arch.evm.cfg import EvmCFG


# Etherem smart contract == EVM bytecode
class EthereumCFG(EvmCFG):
    def __init__(self, bytecode, analysis='dynamic'):
        EvmCFG.__init__(self,
                        bytecode=bytecode,
                        analysis=analysis)



# extract from https://github.com/neo-project/neo-vm

pops = 0
pushes = 0

# fee doc - http://docs.neo.org/en-us/sc/systemfees.html
default = 0.001

_table = {
    # opcode:(mnemonic, immediate_operand_size, pops, pushes, gas, description)
    0x00: ('PUSH0', 0, 0, 1, 0, 'An empty array of bytes is pushed onto the stack.'),
    0x01: ('PUSHBYTES1', 0x01, 0, 0x01, 0, '0x01 byte will be push onto the stack'),
    0x02: ('PUSHBYTES2', 0x02, 0, 0x02, 0, '0x02 bytes will be push onto the stack'),
    0x03: ('PUSHBYTES3', 0x03, 0, 0x03, 0, '0x03 bytes will be push onto the stack'),
    0x04: ('PUSHBYTES4', 0x04, 0, 0x04, 0, '0x04 bytes will be push onto the stack'),
    0x05: ('PUSHBYTES5', 0x05, 0, 0x05, 0, '0x05 bytes will be push onto the stack'),
    0x06: ('PUSHBYTES6', 0x06, 0, 0x06, 0, '0x06 bytes will be push onto the stack'),
    0x07: ('PUSHBYTES7', 0x07, 0, 0x07, 0, '0x07 bytes will be push onto the stack'),
    0x08: ('PUSHBYTES8', 0x08, 0, 0x08, 0, '0x08 bytes will be push onto the stack'),
    0x09: ('PUSHBYTES9', 0x09, 0, 0x09, 0, '0x09 bytes will be push onto the stack'),
    0x0A: ('PUSHBYTES10', 0x0A, 0, 0x0A, 0, '0x0A bytes will be push onto the stack'),
    0x0B: ('PUSHBYTES11', 0x0B, 0, 0x0B, 0, '0x0B bytes will be push onto the stack'),
    0x0C: ('PUSHBYTES12', 0x0C, 0, 0x0C, 0, '0x0C bytes will be push onto the stack'),
    0x0D: ('PUSHBYTES13', 0x0D, 0, 0x0D, 0, '0x0D bytes will be push onto the stack'),
    0x0E: ('PUSHBYTES14', 0x0E, 0, 0x0E, 0, '0x0E bytes will be push onto the stack'),
    0x0F: ('PUSHBYTES15', 0x0F, 0, 0x0F, 0, '0x0F bytes will be push onto the stack'),
    0x10: ('PUSHBYTES16', 0x10, 0, 0x10, 0, '0x10 bytes will be push onto the stack'),
    0x11: ('PUSHBYTES17', 0x11, 0, 0x11, 0, '0x11 bytes will be push onto the stack'),
    0x12: ('PUSHBYTES18', 0x12, 0, 0x12, 0, '0x12 bytes will be push onto the stack'),
    0x13: ('PUSHBYTES19', 0x13, 0, 0x13, 0, '0x13 bytes will be push onto the stack'),
    0x14: ('PUSHBYTES20', 0x14, 0, 0x14, 0, '0x14 bytes will be push onto the stack'),
    0x15: ('PUSHBYTES21', 0x15, 0, 0x15, 0, '0x15 bytes will be push onto the stack'),
    0x16: ('PUSHBYTES22', 0x16, 0, 0x16, 0, '0x16 bytes will be push onto the stack'),
    0x17: ('PUSHBYTES23', 0x17, 0, 0x17, 0, '0x17 bytes will be push onto the stack'),
    0x18: ('PUSHBYTES24', 0x18, 0, 0x18, 0, '0x18 bytes will be push onto the stack'),
    0x19: ('PUSHBYTES25', 0x19, 0, 0x19, 0, '0x19 bytes will be push onto the stack'),
    0x1A: ('PUSHBYTES26', 0x1A, 0, 0x1A, 0, '0x1A bytes will be push onto the stack'),
    0x1B: ('PUSHBYTES27', 0x1B, 0, 0x1B, 0, '0x1B bytes will be push onto the stack'),
    0x1C: ('PUSHBYTES28', 0x1C, 0, 0x1C, 0, '0x1C bytes will be push onto the stack'),
    0x1D: ('PUSHBYTES29', 0x1D, 0, 0x1D, 0, '0x1D bytes will be push onto the stack'),
    0x1E: ('PUSHBYTES30', 0x1E, 0, 0x1E, 0, '0x1E bytes will be push onto the stack'),
    0x1F: ('PUSHBYTES31', 0x1F, 0, 0x1F, 0, '0x1F bytes will be push onto the stack'),
    0x20: ('PUSHBYTES32', 0x20, 0, 0x20, 0, '0x20 bytes will be push onto the stack'),
    0x21: ('PUSHBYTES33', 0x21, 0, 0x21, 0, '0x21 bytes will be push onto the stack'),
    0x22: ('PUSHBYTES34', 0x22, 0, 0x22, 0, '0x22 bytes will be push onto the stack'),
    0x23: ('PUSHBYTES35', 0x23, 0, 0x23, 0, '0x23 bytes will be push onto the stack'),
    0x24: ('PUSHBYTES36', 0x24, 0, 0x24, 0, '0x24 bytes will be push onto the stack'),
    0x25: ('PUSHBYTES37', 0x25, 0, 0x25, 0, '0x25 bytes will be push onto the stack'),
    0x26: ('PUSHBYTES38', 0x26, 0, 0x26, 0, '0x26 bytes will be push onto the stack'),
    0x27: ('PUSHBYTES39', 0x27, 0, 0x27, 0, '0x27 bytes will be push onto the stack'),
    0x28: ('PUSHBYTES40', 0x28, 0, 0x28, 0, '0x28 bytes will be push onto the stack'),
    0x29: ('PUSHBYTES41', 0x29, 0, 0x29, 0, '0x29 bytes will be push onto the stack'),
    0x2A: ('PUSHBYTES42', 0x2A, 0, 0x2A, 0, '0x2A bytes will be push onto the stack'),
    0x2B: ('PUSHBYTES43', 0x2B, 0, 0x2B, 0, '0x2B bytes will be push onto the stack'),
    0x2C: ('PUSHBYTES44', 0x2C, 0, 0x2C, 0, '0x2C bytes will be push onto the stack'),
    0x2D: ('PUSHBYTES45', 0x2D, 0, 0x2D, 0, '0x2D bytes will be push onto the stack'),
    0x2E: ('PUSHBYTES46', 0x2E, 0, 0x2E, 0, '0x2E bytes will be push onto the stack'),
    0x2F: ('PUSHBYTES47', 0x2F, 0, 0x2F, 0, '0x2F bytes will be push onto the stack'),
    0x30: ('PUSHBYTES48', 0x30, 0, 0x30, 0, '0x30 bytes will be push onto the stack'),
    0x31: ('PUSHBYTES49', 0x31, 0, 0x31, 0, '0x31 bytes will be push onto the stack'),
    0x32: ('PUSHBYTES50', 0x32, 0, 0x32, 0, '0x32 bytes will be push onto the stack'),
    0x33: ('PUSHBYTES51', 0x33, 0, 0x33, 0, '0x33 bytes will be push onto the stack'),
    0x34: ('PUSHBYTES52', 0x34, 0, 0x34, 0, '0x34 bytes will be push onto the stack'),
    0x35: ('PUSHBYTES53', 0x35, 0, 0x35, 0, '0x35 bytes will be push onto the stack'),
    0x36: ('PUSHBYTES54', 0x36, 0, 0x36, 0, '0x36 bytes will be push onto the stack'),
    0x37: ('PUSHBYTES55', 0x37, 0, 0x37, 0, '0x37 bytes will be push onto the stack'),
    0x38: ('PUSHBYTES56', 0x38, 0, 0x38, 0, '0x38 bytes will be push onto the stack'),
    0x39: ('PUSHBYTES57', 0x39, 0, 0x39, 0, '0x39 bytes will be push onto the stack'),
    0x3A: ('PUSHBYTES58', 0x3A, 0, 0x3A, 0, '0x3A bytes will be push onto the stack'),
    0x3B: ('PUSHBYTES59', 0x3B, 0, 0x3B, 0, '0x3B bytes will be push onto the stack'),
    0x3C: ('PUSHBYTES60', 0x3C, 0, 0x3C, 0, '0x3C bytes will be push onto the stack'),
    0x3D: ('PUSHBYTES61', 0x3D, 0, 0x3D, 0, '0x3D bytes will be push onto the stack'),
    0x3E: ('PUSHBYTES62', 0x3E, 0, 0x3E, 0, '0x3E bytes will be push onto the stack'),
    0x3F: ('PUSHBYTES63', 0x3F, 0, 0x3F, 0, '0x3F bytes will be push onto the stack'),
    0x40: ('PUSHBYTES64', 0x40, 0, 0x40, 0, '0x40 bytes will be push onto the stack'),
    0x41: ('PUSHBYTES65', 0x41, 0, 0x41, 0, '0x41 bytes will be push onto the stack'),
    0x42: ('PUSHBYTES66', 0x42, 0, 0x42, 0, '0x42 bytes will be push onto the stack'),
    0x43: ('PUSHBYTES67', 0x43, 0, 0x43, 0, '0x43 bytes will be push onto the stack'),
    0x44: ('PUSHBYTES68', 0x44, 0, 0x44, 0, '0x44 bytes will be push onto the stack'),
    0x45: ('PUSHBYTES69', 0x45, 0, 0x45, 0, '0x45 bytes will be push onto the stack'),
    0x46: ('PUSHBYTES70', 0x46, 0, 0x46, 0, '0x46 bytes will be push onto the stack'),
    0x47: ('PUSHBYTES71', 0x47, 0, 0x47, 0, '0x47 bytes will be push onto the stack'),
    0x48: ('PUSHBYTES72', 0x48, 0, 0x48, 0, '0x48 bytes will be push onto the stack'),
    0x49: ('PUSHBYTES73', 0x49, 0, 0x49, 0, '0x49 bytes will be push onto the stack'),
    0x4A: ('PUSHBYTES74', 0x4A, 0, 0x4A, 0, '0x4A bytes will be push onto the stack'),
    0x4B: ('PUSHBYTES75', 0x4B, 0, 0x4B, 0, '0x4B datas will be push onto the stack'),
    0x4C: ('PUSHDATA1', 1, 0, 1, 0, 'The next byte contains the number of bytes to be pushed onto the stack.'),
    0x4D: ('PUSHDATA2', 2, 0, 2, 0, 'The next two bytes contain the number of bytes to be pushed onto the stack.'),
    0x4E: ('PUSHDATA4', 4, 0, 4, 0, 'The next four bytes contain the number of bytes to be pushed onto the stack.'),
    0x4F: ('PUSHM1', 0, 0, 1, 0, 'The number -1 is pushed onto the stack.'),
    0x51: ('PUSH1', 0, 0, 1, 0, 'The number 1 is pushed onto the stack.'),
    0x52: ('PUSH2', 0, 0, 1, 0, 'The number 2 is pushed onto the stack.'),
    0x53: ('PUSH3', 0, 0, 1, 0, 'The number 3 is pushed onto the stack.'),
    0x54: ('PUSH4', 0, 0, 1, 0, 'The number 4 is pushed onto the stack.'),
    0x55: ('PUSH5', 0, 0, 1, 0, 'The number 5 is pushed onto the stack.'),
    0x56: ('PUSH6', 0, 0, 1, 0, 'The number 6 is pushed onto the stack.'),
    0x57: ('PUSH7', 0, 0, 1, 0, 'The number 7 is pushed onto the stack.'),
    0x58: ('PUSH8', 0, 0, 1, 0, 'The number 8 is pushed onto the stack.'),
    0x59: ('PUSH9', 0, 0, 1, 0, 'The number 9 is pushed onto the stack.'),
    0x5A: ('PUSH10', 0, 0, 1, 0, 'The number 10 is pushed onto the stack.'),
    0x5B: ('PUSH11', 0, 0, 1, 0, 'The number 11 is pushed onto the stack.'),
    0x5C: ('PUSH12', 0, 0, 1, 0, 'The number 12 is pushed onto the stack.'),
    0x5D: ('PUSH13', 0, 0, 1, 0, 'The number 13 is pushed onto the stack.'),
    0x5E: ('PUSH14', 0, 0, 1, 0, 'The number 14 is pushed onto the stack.'),
    0x5F: ('PUSH15', 0, 0, 1, 0, 'The number 15 is pushed onto the stack.'),
    0x60: ('PUSH16', 0, 0, 1, 0, 'The number 16 is pushed onto the stack.'),

    # Flow control
    0x61: ('NOP', 0, 0, 0, 0, 'Does nothing.'),
    0x62: ('JMP', 2, 0, 0, default, 'JUMP'),
    0x63: ('JMPIF', 2, 1, 0, default, 'CONDITIONNAL JUMP'),
    0x64: ('JMPIFNOT', 2, 1, 0, default, 'NEGATIVE CONDITIONNAL JUMP'),
    0x65: ('CALL', 2, 0, pushes, default, 'CALL FUNCTION'),
    0x66: ('RET', 0, 0, 0, default, 'RET FUNCTION'),
    0x67: ('APPCALL', 20, pops, pushes, 0.01, 'APPCALL'),
    0x68: ('SYSCALL', 1, pops, pushes, default, 'SYSCALL'),
    0x69: ('TAILCALL', 20, pops, pushes, 0.01, 'TAILCALL'),

    # Stack
    0x6A: ('DUPFROMALTSTACK', 0, 0, 1, default, ''),
    0x6B: ('TOALTSTACK', 0, 1, 1, default, 'Puts the input onto the top of the alt stack. Removes it from the main stack.'),
    0x6C: ('FROMALTSTACK', 0, 1, 1, default, 'Puts the input onto the top of the main stack. Removes it from the alt stack.'),
    0x6D: ('XDROP', 0, pops, pushes, default, ''),
    0x72: ('XSWAP', 0, 1, 0, default, ''),
    0x73: ('XTUCK', 0, 1, 1, default, ''),
    0x74: ('DEPTH', 0, 0, 1, default, 'Puts the number of stack items onto the stack.'),
    0x75: ('DROP', 0, 1, 0, default, 'Removes the top stack item.'),
    0x76: ('DUP', 0, 0, 1, default, 'Duplicates the top stack item.'),
    0x77: ('NIP', 0, 2, 1, default, 'Removes the second-to-top stack item.'),
    0x78: ('OVER', 0, 1, 2, default, 'Copies the second-to-top stack item to the top.'),
    0x79: ('PICK', 0, 1, 1, default, 'The item n back in the stack is copied to the top.'),
    0x7A: ('ROLL', 0, pops, pushes, default, 'The item n back in the stack is moved to the top.'),
    0x7B: ('ROT', 0, 3, 3, default, 'The top three items on the stack are rotated to the left.'),
    0x7C: ('SWAP', 0, 2, 2, default, 'The top two items on the stack are swapped.'),
    0x7D: ('TUCK', 0, 2, 3, default, 'The item at the top of the stack is copied and inserted before the second-to-top item.'),

    # Splice
    0x7E: ('CAT', 0, 2, 1, default, 'Concatenates two strings.'),
    0x7F: ('SUBSTR', 0, 3, 1, default, 'Returns a section of a string.'),
    0x80: ('LEFT', 0, 2, 1, default, 'Keeps only characters left of the specified point in a string.'),
    0x81: ('RIGHT', 0, 2, 1, default, 'Keeps only characters right of the specified point in a string.'),
    0x82: ('SIZE', 0, 1, 1, default, 'Returns the length of the input string.'),

    # Bitwise logic
    0x83: ('INVERT', 0, 1, 1, default, 'Flips all of the bits in the input.'),
    0x84: ('AND', 0, 2, 1, default, 'Boolean and between each bit in the inputs.'),
    0x85: ('OR', 0, 2, 1, default, 'Boolean or between each bit in the inputs.'),
    0x86: ('XOR', 0, 2, 1, default, 'Boolean exclusive or between each bit in the inputs.'),
    0x87: ('EQUAL', 0, 2, 1, default, 'Returns 1 if the inputs are exactly equal, 0 otherwise.'),
    0x88: ('OP_EQUALVERIFY', 0, pops, pushes, default, 'Same as OP_EQUAL, but runs OP_VERIFY afterward.'),
    0x89: ('OP_RESERVED1', 0, pops, pushes, default, 'Transaction is invalid unless occuring in an unexecuted OP_IF branch'),
    0x8A: ('OP_RESERVED2', 0, pops, pushes, default, 'Transaction is invalid unless occuring in an unexecuted OP_IF branch'),

    # Arithmetic
    # Note: Arithmetic inputs are limited to signed 32-bit integers, but may overflow their output.
    0x8B: ('INC', 0, 1, 1, default, '1 is added to the input.'),
    0x8C: ('DEC', 0, 1, 1, default, '1 is subtracted from the input.'),
    0x8D: ('SIGN', 0, 1, 1, default, ''),
    0x8F: ('NEGATE', 0, 1, 1, default, 'The sign of the input is flipped.'),
    0x90: ('ABS', 0, 1, 1, default, 'The input is made positive.'),
    0x91: ('NOT', 0, 1, 1, default, 'If the input is 0 or 1, it is flipped. Otherwise the output will be 0.'),
    0x92: ('NZ', 0, 1, 1, default, 'Returns 0 if the input is 0. 1 otherwise.'),
    0x93: ('ADD', 0, pops, pushes, default, 'a is added to b.'),
    0x94: ('SUB', 0, pops, pushes, default, 'b is subtracted from a.'),
    0x95: ('MUL', 0, pops, pushes, default, 'a is multiplied by b.'),
    0x96: ('DIV', 0, pops, pushes, default, 'a is divided by b.'),
    0x97: ('MOD', 0, pops, pushes, default, 'Returns the remainder after dividing a by b.'),
    0x98: ('SHL', 0, pops, pushes, default, 'Shifts a left b bits, preserving sign.'),
    0x99: ('SHR', 0, pops, pushes, default, 'Shifts a right b bits, preserving sign.'),
    0x9A: ('BOOLAND', 0, pops, pushes, default, 'If both a and b are not 0, the output is 1. Otherwise 0.'),
    0x9B: ('BOOLOR', 0, pops, pushes, default, 'If a or b is not 0, the output is 1. Otherwise 0.'),
    0x9C: ('NUMEQUAL', 0, pops, pushes, default, 'Returns 1 if the numbers are equal, 0 otherwise.'),
    0x9E: ('NUMNOTEQUAL', 0, pops, pushes, default, 'Returns 1 if the numbers are not equal, 0 otherwise.'),
    0x9F: ('LT', 0, pops, pushes, default, 'Returns 1 if a is less than b, 0 otherwise.'),
    0xA0: ('GT', 0, pops, pushes, default, 'Returns 1 if a is greater than b, 0 otherwise.'),
    0xA1: ('LTE', 0, pops, pushes, default, 'Returns 1 if a is less than or equal to b, 0 otherwise.'),
    0xA2: ('GTE', 0, pops, pushes, default, 'Returns 1 if a is greater than or equal to b, 0 otherwise.'),
    0xA3: ('MIN', 0, pops, pushes, default, 'Returns the smaller of a and b.'),
    0xA4: ('MAX', 0, pops, pushes, default, 'Returns the larger of a and b.'),
    0xA5: ('WITHIN', 0, pops, pushes, default, 'Returns 1 if x is within the specified range (left-inclusive), 0 otherwise.'),


    # Crypto
    #0xA6: ('RIPEMD160', 0, pops, pushes, default, 'The input is hashed using RIPEMD-160.'),
    0xA7: ('SHA1', 0, pops, pushes, 0.01, 'The input is hashed using SHA-1.'),
    0xA8: ('SHA256', 0, pops, pushes, 0.01, 'The input is hashed using SHA-256.'),
    0xA9: ('HASH160', 0, pops, pushes, 0.02, ''),
    0xAA: ('HASH256', 0, pops, pushes, 0.02, ''),
    0xAC: ('CHECKSIG', 0, pops, pushes, 0.1, ''),
    0xAD: ('VERIFY', 0, pops, pushes, 0.1, ''),
    0xAE: ('CHECKMULTISIG', 0, pops, pushes, 0.1, ''),


    # Array
    0xC0: ('ARRAYSIZE', 0, pops, pushes, default, ''),
    0xC1: ('PACK', 0, pops, pushes, default, ''),
    0xC2: ('UNPACK', 0, pops, pushes, default, ''),
    0xC3: ('PICKITEM', 0, pops, pushes, default, ''),
    0xC4: ('SETITEM', 0, pops, pushes, default, ''),
    0xC5: ('NEWARRAY', 0, pops, pushes, default, 'Used as reference type'),
    0xC6: ('NEWSTRUCT', 0, pops, pushes, default, 'Used as reference value '),
    0xC7: ('NEWMAP', 0, pops, pushes, default, ''),
    0xC8: ('APPEND', 0, pops, pushes, default, ''),
    0xC9: ('REVERSE', 0, pops, pushes, default, ''),
    0xCA: ('REMOVE', 0, pops, pushes, default, ''),
    0xCB: ('HASKEY', 0, pops, pushes, default, ''),
    0xCC: ('KEYS', 0, pops, pushes, default, ''),
    0xCD: ('VALUES', 0, pops, pushes, default, ''),

    # Exceptions
    0xF0: ('THROW', 0, 0, 0, default, ''),
    0xF1: ('THROWIFNOT', 0, 1, 0, default, '')
}


class Avm(object):
    """Bytecode for NEO VM."""

    def __init__(self):
        self.table = _table
        self.reverse_table = self._get_reverse_table()

    def _get_reverse_table(self):
        """Build an internal table used in the assembler."""
        reverse_table = {}
        for (opcode, (mnemonic, immediate_operand_size,
                      pops, pushes, gas, description)) in self.table.items():
            reverse_table[mnemonic] = opcode, mnemonic, immediate_operand_size, \
                pops, pushes, gas, description
        return reverse_table



from wasm.modtypes import (TypeSection,
                           ImportSection,
                           FunctionSection,
                           TableSection,
                           MemorySection,
                           GlobalSection,
                           ExportSection,
                           StartSection,
                           ElementSection,
                           CodeSection,
                           DataSection)

from octopus.arch.wasm.constant import LANG_TYPE, KIND_TYPE
from octopus.arch.wasm.format import (format_kind_function,
                                      format_kind_table,
                                      format_kind_memory,
                                      format_kind_global)

from octopus.arch.wasm.decode import decode_module

from octopus.core.utils import bytecode_to_bytes
# from wasm.decode import decode_module
import io
import logging
log = logging.getLogger(__name__)
log.setLevel(level=logging.INFO)


class WasmModuleAnalyzer(object):
    '''Analyze and extract informations from wasm module'''

    def __init__(self, module_bytecode, analysis=True):
        self.module_bytecode = bytecode_to_bytes(module_bytecode)

        self.magic = None
        self.version = None
        self.types = list()
        self.imports_all = list()
        self.imports_func = list()
        self.func_types = list()
        self.tables = list()
        self.memories = list()
        self.globals = list()
        self.exports = list()
        self.start = None
        self.elements = list()
        self.codes = list()
        self.datas = list()
        self.names = list()
        self.customs = list()
        self.func_prototypes = list()
        # self.strings = list() - TODO

        if analysis:
            self.analyze()

    def attributes_reset(self):
        self.magic = None
        self.version = None
        self.types = list()
        self.imports_all = list()
        self.imports_func = list()
        self.func_types = list()
        self.tables = list()
        self.memories = list()
        self.globals = list()
        self.exports = list()
        self.start = None
        self.elements = list()
        self.codes = list()
        self.datas = list()
        self.names = list()
        self.customs = list()
        self.func_prototypes = list()

    def show(self):
        """Return dict with WasmModuleAnalyzer attributes"""
        return {'magic': self.magic,
                'version': self.version,
                'types': self.types,
                'imports_all': self.imports_all,
                'imports_func': self.imports_func,
                'func_types': self.func_types,
                'tables': self.tables,
                'memories': self.memories,
                'globals': self.globals,
                'exports': self.exports,
                'start': self.start,
                'elements': self.elements,
                'length codes': len(self.codes),
                'datas': self.datas}

    def get_section(self, section_type):
        mod_iter = iter(decode_module(self.module_bytecode))
        _, _ = next(mod_iter)
        sections = list(mod_iter)

        # iterate over all section
        for cur_sec, cur_sec_data in sections:
            sec = cur_sec_data.get_decoder_meta()['types']['payload']

            if isinstance(sec, section_type):
                return cur_sec_data
        return None

    def __decode_header(self, header, h_data):
        """Decode wasm header
        Return tuple (magic, version) of wasm module header
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#high-level-structure
        """
        magic = \
            h_data.magic.to_bytes(header.magic.byte_size, 'little')
        version = \
            h_data.version.to_bytes(header.version.byte_size, 'little')
        return (magic, version)

    def __decode_type_section(self, type_section):
        """Decode wasm type section
        Return a list of tuple (param_str, return_str)
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#type-section
        """
        type_list = []

        for idx, entry in enumerate(type_section.payload.entries):
            param_str = ''
            return_str = ''

            param_str += ' '.join([LANG_TYPE.get(_x) for _x in entry.param_types])
            if entry.return_type:
                return_str = '%s' % LANG_TYPE.get(entry.return_type)

            type_list.append((param_str, return_str))
        return type_list

    def __decode_import_section(self, import_section):
        """Decode import section to tuple of list
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#import-section
        """
        entries = import_section.payload.entries
        import_list = []
        import_func_list = []

        for idx, entry in enumerate(entries):
            #            for encoding in ('utf-8', 'utf-16-be'):
            #                value = str(v)
            #                try:
            #                    value = v.decode(encoding)
            #                    break
            #                except UnicodeDecodeError:
            #                    value = str(v)
            try:
                module_str = entry.module_str.tobytes().decode('utf-8')
            except UnicodeDecodeError:
                module_str = entry.module_str.tobytes()
            try:
                field_str = entry.field_str.tobytes().decode('utf-8')
            except UnicodeDecodeError:
                field_str = entry.field_str.tobytes()

            log.debug('%s %s', module_str, field_str)
            kind_type = KIND_TYPE.get(entry.kind)

            if kind_type == 'function':
                f_type = format_kind_function(entry.type.type)
                import_list.append((entry.kind, module_str, field_str, f_type))
                # add also the info into the specific import function list
                import_func_list.append((module_str, field_str, f_type))
            elif kind_type == 'table':
                tabl = format_kind_table(entry.type.element_type,
                                         entry.type.limits.flags,
                                         entry.type.limits.initial,
                                         entry.type.limits.maximum)
                import_list.append((entry.kind, module_str, field_str, tabl))

            elif kind_type == 'memory':
                mem = format_kind_memory(entry.type.limits.flags,
                                         entry.type.limits.initial,
                                         entry.type.limits.maximum)
                import_list.append((entry.kind, module_str, field_str, mem))

            elif kind_type == 'global':
                gbl = format_kind_global(entry.type.content_type,
                                         entry.type.mutability)
                import_list.append((entry.kind, module_str, field_str, gbl))
            else:
                log.error('unknown %d %s %s', entry.kind,
                          module_str, field_str)
        return (import_list, import_func_list)

    def __decode_function_section(self, function_section):
        """Decode function section
        The function section declares the signatures of all functions in the module
        Return list of indices (int)
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#function-section
        """
        return function_section.payload.types

    def __decode_table_section(self, table_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#table-section
        """
        # on the MVP, table size == 1
        entries = table_section.payload.entries
        table_list = []

        for idx, entry in enumerate(entries):
            element_type = entry.element_type
            flags = entry.limits.flags
            initial = entry.limits.initial
            maximum = entry.limits.maximum

            fmt = format_kind_table(element_type,
                                    flags,
                                    initial,
                                    maximum)
            table_list.append(fmt)
        return table_list

    def __decode_memory_section(self, memory_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#memory-section
        """
        # on the MVP, memory size == 1
        memory_l = list()
        entries = memory_section.payload.entries

        for idx, entry in enumerate(entries):

            flags = entry.limits.flags
            initial = entry.limits.initial
            maximum = entry.limits.maximum

            fmt = format_kind_memory(flags,
                                     initial,
                                     maximum)

            memory_l.append(fmt)
        return memory_l

    def __decode_global_section(self, global_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#global-section
        """
        globals_l = list()
        for entry in global_section.payload.globals:
            fmt = format_kind_global(entry.type.mutability,
                                     entry.type.content_type)
            globals_l.append(fmt)
        return globals_l

    def __decode_export_section(self, export_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#export-section
        """
        entries = export_section.payload.entries
        export_list = []

        for idx, entry in enumerate(entries):
            # field_str == function_name
            try:
                field_str = entry.field_str.tobytes().decode('utf-8')
            except UnicodeDecodeError:
                field_str = entry.field_str.tobytes()
            kind = entry.kind
            index = entry.index

            fmt = {'field_str': field_str,
                   'kind': kind,
                   'index': index}
            export_list.append(fmt)
        return export_list

    def __decode_start_section(self, start_section):
        return start_section.payload.index

    def __decode_element_section(self, element_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#element-section
        """
        entries = element_section.payload.entries
        element_list = []

        for idx, entry in enumerate(entries):

            fmt = {'index': entry.index,
                   'offset': entry.offset,
                   'elems': entry.elems}

            element_list.append(fmt)
        return element_list

    def __decode_code_section(self, code_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#code-section
        """
        bodies = code_section.payload.bodies
        code_list = []

        for idx, entry in enumerate(bodies):
            code_raw = entry.code.tobytes()
            code_list.append(code_raw)
        return code_list

    def __decode_data_section(self, data_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#data-section
        """
        entries = data_section.payload.entries
        data_list = []

        for idx, entry in enumerate(entries):
            data = entry.data.tobytes()

            fmt = {'index': entry.index,
                   'offset': entry.offset,
                   'size': entry.size,
                   'data': data}

            data_list.append(fmt)
        return data_list

    def __decode_name_section(self, name_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#name-section
        """
        payload = name_section.payload.tobytes()
        total = 0
        names_list = list()

        f = io.BytesIO(payload)
        count = int.from_bytes(f.read(1), byteorder='big')
        total += 1

        while total < len(payload) - 1:
            index = int.from_bytes(f.read(1), byteorder='big')
            total += 1
            name_len = int.from_bytes(f.read(1), byteorder='big')
            total += 1
            name_str = f.read(name_len)
            total += name_len
            names_list.append((index, name_len, name_str))
        f.close()
        return names_list

    def __decode_unknown_section(self, unknown_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#high-level-structure
        """
        sec_name = unknown_section.name.tobytes()
        payload = unknown_section.payload.tobytes()
        return (sec_name, payload)

    def _create_ordered_list(self):
        """create ordered list of functions"""

        func_prototypes = list()

        # get imported functions
        for _, name, type_idx in self.imports_func:
            _param, _return = self.types[type_idx]
            func_prototypes.append((name, _param, _return))

        # get all internal functions
        for idx, code in enumerate(self.codes):
            _param, _return = self.types[self.func_types[idx]]
            real_index = len(self.imports_func) + idx
            name = '$func%d' % real_index

            # if exported function - overwrite name
            for x in self.exports:
                if x.get('index') == real_index and x.get('kind') == 0:
                    name = x.get('field_str')

            # TODO: need to test
            if real_index == self.start:
                name = '* ' + name
            func_prototypes.append((name, _param, _return))
        return func_prototypes

    def analyze(self):
        """analyse the complete module & extract informations """
        # src: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md
        # custom     0   name, .debug_str, ...
        # Type       1   Function signature declarations
        # Import     2   Import declarations
        # Function   3   Function declarations
        # Table      4   Indirect function table and other tables
        # Memory     5   Memory attributes
        # Global     6   Global declarations
        # Export     7   Exports
        # Start      8   Start function declaration
        # Element    9   Elements section
        # Code       10  Function bodies (code)
        # Data       11  Data segments

        # reset attributes
        self.attributes_reset()

        mod_iter = iter(decode_module(self.module_bytecode))
        # decode header version - usefull in the future (multiple versions)
        header, header_data = next(mod_iter)
        self.magic, self.version = self.__decode_header(header, header_data)

        #
        # Wasm sections
        #
        sections = list(mod_iter)

        for cur_sec, cur_sec_data in sections:
            sec = cur_sec_data.get_decoder_meta()['types']['payload']

            if isinstance(sec, TypeSection):
                self.types = self.__decode_type_section(cur_sec_data)
            elif isinstance(sec, ImportSection):
                self.imports_all, self.imports_func = \
                    self.__decode_import_section(cur_sec_data)
            elif isinstance(sec, FunctionSection):
                self.func_types = self.__decode_function_section(cur_sec_data)
            elif isinstance(sec, TableSection):
                self.tables = self.__decode_table_section(cur_sec_data)
            elif isinstance(sec, MemorySection):
                self.memories = self.__decode_memory_section(cur_sec_data)
            elif isinstance(sec, GlobalSection):
                # TODO not analyzed
                self.globals = self.__decode_global_section(cur_sec_data)
            elif isinstance(sec, ExportSection):
                self.exports = self.__decode_export_section(cur_sec_data)
            elif isinstance(sec, StartSection):
                # TODO not analyzed
                self.start = self.__decode_start_section(cur_sec_data)
            elif isinstance(sec, ElementSection):
                self.elements = self.__decode_element_section(cur_sec_data)
            elif isinstance(sec, CodeSection):
                self.codes = self.__decode_code_section(cur_sec_data)
            elif isinstance(sec, DataSection):
                self.datas = self.__decode_data_section(cur_sec_data)
            else:
                # name section
                if cur_sec_data.id == 0 and cur_sec_data.name.tobytes() == b'name':
                    self.names = self.__decode_name_section(cur_sec_data)
                else:
                    # TODO - handle properly .debug_str section
                    self.customs.append(self.__decode_unknown_section(cur_sec_data))

        # create ordered list of functions
        self.func_prototypes = self._create_ordered_list()
        return self.show()



from octopus.arch.wasm.constant import LANG_TYPE


def format_func_name(name, param_str, return_str):
    result = '{} '.format(return_str) if return_str else ''
    return ('{}{}({})'.format(result, name, param_str))


def format_bb_name(function_id, offset):
    return ('block_%x_%x' % (function_id, offset))


def format_kind_function(f_type):
    return f_type


def format_kind_table(element_type, flags, initial, maximum):
    return {'element_type': LANG_TYPE.get(element_type),
            'limits_flags': flags,
            'limits_initial': initial,
            'limits_maximum': maximum}


def format_kind_memory(flags, initial, maximum):
    return {'limits_flags': flags,
            'limits_initial': initial,
            'limits_maximum': maximum}


def format_kind_global(content_type, mutability):
    return (content_type, mutability)



import warnings

#from ethereum import utils
#from ethereum.abi import encode_abi, decode_abi

from octopus.platforms.ETH.constants import DEFAULT_GAS_PER_TX, DEFAULT_GAS_PRICE, BLOCK_TAGS, BLOCK_TAG_LATEST
from octopus.platforms.ETH.util import hex_to_dec, clean_hex, validate_block

from octopus.engine.explorer import Explorer
"""
This code is adapted from: ethjsonrpc
https://github.com/ConsenSys/ethjsonrpc
"""

GETH_DEFAULT_RPC_PORT = 8545
ETH_DEFAULT_RPC_PORT = 8545
PARITY_DEFAULT_RPC_PORT = 8545
PYETHAPP_DEFAULT_RPC_PORT = 4000

INFURA_MAINNET = "mainnet.infura.io/"
INFURA_ROPSTEN = "ropsten.infura.io/"
INFURA_INFURANET = "infuranet.infura.io/"
INFURA_KOVAN = "kovan.infura.io/"
INFURA_RINKEBY = "rinkeby.infura.io/"
INFURA_RPC_PORT = 8545


class EthereumExplorerRPC(Explorer):
    """
    Ethereum JSON-RPC client class
    """
    def __init__(self, host='localhost', port=GETH_DEFAULT_RPC_PORT, tls=False, max_retries=3):
        Explorer.__init__(self, host=host, port=port, tls=tls, max_retries=max_retries)

    def call(self, method, params=None, jsonrpc='2.0', _id=1):
        return super().call(method, params, jsonrpc, _id)

    '''
    def _encode_function(self, signature, param_values):

        prefix = utils.big_endian_to_int(utils.sha3(signature)[:4])

        if signature.find('(') == -1:
            raise RuntimeError('Invalid function signature. Missing "(" and/or ")"...')

        if signature.find(')') - signature.find('(') == 1:
            return utils.encode_int(prefix)

        types = signature[signature.find('(') + 1: signature.find(')')].split(',')
        encoded_params = encode_abi(types, param_values)
        return utils.zpad(utils.encode_int(prefix), 4) + encoded_params
    '''

    #######################
    # HIGHT-LEVEL METHODS #
    #######################

    def get_transaction(self, transaction_id, verbosity=None):
        """ Return transaction informations

        .. seealso::
            :method:`eth_getTransactionByHash`
        """
        return self.eth_getTransactionByHash(transaction_id)

    def get_block_by_number(self, block_number):
        """ Return block information using given block number

        .. seealso::
            :method:`eth_getBlockByNumber`
        """
        return self.eth_getBlockByNumber(block_number)

    def get_block_by_hash(self, block_hash):
        """ Return block information using given block hash

        .. seealso::
            :method:`eth_getBlockByHash`
        """
        return self.eth_getBlockByHash(block_hash)

    def decode_tx(self, transaction_id):
        """ Return dict with important information about
            the given transaction
        """
        tx_data = self.eth_getTransactionByHash(transaction_id)
        return tx_data
        #TODO


    ##########################
    # HIGHT-LEVEL METHODS #2 #
    ##########################

    def transfer(self, from_address, to_address, amount):
        """Send wei from one address to another

        TODO

        .. note:: https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_sendtransaction
        .. seealso:: :method:`eth_sendTransaction`
        """
        return self.eth_sendTransaction(from_address=from_address, to_address=to_address, value=amount)

    def create_contract(self, from_, code, gas, sig=None, args=None):
        """
        Create a contract on the blockchain from compiled EVM code. Returns the
        transaction hash.
        """
        '''
        from_ = from_ or self.eth_coinbase()
        if sig is not None and args is not None:
            types = sig[sig.find('(') + 1: sig.find(')')].split(',')
            encoded_params = encode_abi(types, args)
            code += encoded_params.encode('hex')
        return self.eth_sendTransaction(from_address=from_, gas=gas, data=code)
        '''
        return NotImplementedError()

    def get_contract_address(self, tx):
        """
        Get the address for a contract from the transaction that created it
        """
        receipt = self.eth_getTransactionReceipt(tx)
        return receipt['contractAddress']

    def call_without_transaction(self, address, sig, args, result_types):
        """
        Call a contract function on the RPC server, without sending a
        transaction (useful for reading data)
        """
        '''
        data = self._encode_function(sig, args)
        data_hex = data.encode('hex')
        response = self.eth_call(to_address=address, data=data_hex)
        return decode_abi(result_types, response[2:].decode('hex'))
        '''
        return NotImplementedError()

    def call_with_transaction(self, from_, address, sig, args, gas=None, gas_price=None, value=None):
        """
        Call a contract function by sending a transaction (useful for storing
        data)
        """
        '''
        gas = gas or DEFAULT_GAS_PER_TX
        gas_price = gas_price or DEFAULT_GAS_PRICE
        data = self._encode_function(sig, args)
        data_hex = data.encode('hex')
        return self.eth_sendTransaction(from_address=from_, to_address=address, data=data_hex, gas=gas,
                                        gas_price=gas_price, value=value)
        '''
        return NotImplementedError()
    ####################
    # JSON-RPC METHODS #
    ####################

    # ressources :
    # * https://github.com/ethereum/wiki/wiki/JSON-RPC
    # *
    # *

    def web3_clientVersion(self):
        """ Returns the current client version.

        :return: The current client version
        :rtype: str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.web3_clientVersion()
        'Geth/v1.8.0-unstable-952482d5/linux-amd64/go1.9.2'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#web3_clientversion
        .. todo::
            TESTED
        """

        return self.call('web3_clientVersion')

    def web3_sha3(self, data):
        """ Returns Keccak-256 (not the standardized SHA3-256) of the given data.

        :param data: the data to convert into a SHA3 hash
        :type data: hex string
        :return: The SHA3 result of the given string.
        :rtype: hex string

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.web3_sha3('0x' + b'hello world'.hex())
        '0x47173285a8d7341e5e972fc677286384f802f8ef42a5ec5f03bbfa254cb01fad'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#web3_sha3
        .. todo::
            TESTED
        """
        #data = str(data).encode('hex')
        return self.call('web3_sha3', [data])

    def net_version(self):
        """ Returns the current network id.

        :return: The current network id.
            "1": Ethereum Mainnet
            "2": Morden Testnet (deprecated)
            "3": Ropsten Testnet
            "4": Rinkeby Testnet
            "42": Kovan Testnet
        :rtype: str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.net_version()
        '1'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#net_version
        .. todo::
            TESTED
        """
        return self.call('net_version')

    def net_listening(self):
        """ Returns true if client is actively listening for network connections.

        :return: The current network id.
            true when listening otherwise false.
        :rtype: Boolean

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.net_listening()
        True


        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#net_listening
        .. todo::
            TESTED
        """
        return self.call('net_listening')

    def net_peerCount(self):
        """ Returns number of peers currently connected to the client.

        :return: integer of the number of connected peers.
        :rtype: int

        :Example:
        >>> explorer = EthereumExplorerRPC()
        >>> explorer.net_peerCount()
        25

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#net_peercount
        .. todo::
            TESTED
        """
        return hex_to_dec(self.call('net_peerCount'))

    def eth_protocolVersion(self):
        """ Returns the current ethereum protocol version.

        :return: The current ethereum protocol version
        :rtype: hex str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_protocolVersion()
        '0x3f'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_protocolversion
        .. todo::
            TESTED
        """
        return self.call('eth_protocolVersion')

    def eth_syncing(self):
        """ Returns an object with data about the sync status or false.

        :return: An object with sync status data or FALSE, when not syncing
            startingBlock: QUANTITY - The block at which the import started (will only be reset, after the sync reached his head)
            currentBlock: QUANTITY - The current block, same as eth_blockNumber
            highestBlock: QUANTITY - The estimated highest block
        :rtype: Boolean or object

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_syncing()
        False

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_syncing
        .. todo::
            TESTED
        """
        return self.call('eth_syncing')

    def eth_coinbase(self):
        """ Returns the client coinbase address.

        :return: 20 bytes - the current coinbase address.
        :rtype: hex str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_coinbase()
        '0x407d73d8a49eeb85d32cf465507dd71d507100c1'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_coinbase
        .. todo::
            TESTED
        """
        return self.call('eth_coinbase')

    def eth_mining(self):
        """ Returns true if client is actively mining new blocks.

        :return:  returns true of the client is mining, otherwise false
        :rtype: Boolean

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_mining()
        False

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_mining
        .. todo::
            TESTED
        """
        return self.call('eth_mining')

    def eth_hashrate(self):
        """ Returns the number of hashes per second that the node is mining with.

        :return: number of hashes per second.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_hashrate()
        0

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_hashrate
        .. todo::
            TESTED
        """

        return hex_to_dec(self.call('eth_hashrate'))

    def eth_gasPrice(self):
        """ Returns the current price per gas in wei.

        :return: integer of the current gas price in wei.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_gasPrice()
        4000000000

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gasprice
        .. todo::
            TESTED
        """
        return hex_to_dec(self.call('eth_gasPrice'))

    def eth_accounts(self):
        """ Returns a list of addresses owned by client.

        :return: 20 Bytes - addresses owned by the client.
        :rtype: list

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_accounts()
        ["0x407d73d8a49eeb85d32cf465507dd71d507100c1"]

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_accounts
        .. todo::
            TESTED
        """
        return self.call('eth_accounts')

    def eth_blockNumber(self):
        """ Returns the number of most recent block.

        :return: integer of the current block number the client is on.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_blockNumber()
        5100196

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_blocknumber
        .. todo::
            TESTED
        """
        return hex_to_dec(self.call('eth_blockNumber'))

    def eth_getBalance(self, address=None, block=BLOCK_TAG_LATEST):
        """ Returns the balance of the account of given address.

        :param address: 20 Bytes - address to check for balance.
        :type address: str
        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :return:  integer of the current balance in wei.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getBalance("0x956b6B7454884b734B29A8115F045a95179ea00C")
        17410594678300000000

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getbalance
        .. todo::
            TESTED

        """
        address = address or self.eth_coinbase()
        block = validate_block(block)
        v = hex_to_dec(self.call('eth_getBalance', [address, block]))
        return (v if v else 0)

    def eth_getStorageAt(self, address=None, position=0, block=BLOCK_TAG_LATEST):
        """ Returns the value from a storage position at a given address.

        :param address: 20 Bytes - address to check for balance.
        :type address: str
        :param address: (optionnal) integer of the position in the storage. default is 0
        :type address: int
        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :return:  the value at this storage position.
        :rtype: str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getStorageAt("0x295a70b2de5e3953354a6a8344e616ed314d7251", 0, "latest")
        '0x0000000000000000000000000000000000000000000000000000000000000000'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getstorageat
        .. todo::
            TESTED
        """
        block = validate_block(block)
        return self.call('eth_getStorageAt', [address, hex(position), block])

    def eth_getTransactionCount(self, address, block=BLOCK_TAG_LATEST):
        """ Returns the number of transactions sent from an address.

        :param address: 20 Bytes - address.
        :type address: str
        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :return: integer of the number of transactions send from this address.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getTransactionCount("0x956b6B7454884b734B29A8115F045a95179ea00C")
        12891

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gettransactioncount
        .. todo::
            TESTED
        """
        block = validate_block(block)
        return hex_to_dec(self.call('eth_getTransactionCount', [address, block]))

    def eth_getBlockTransactionCountByHash(self, block_hash):
        """ Returns the number of transactions in a block from a block matching the given block hash.

        :param block_hash:  32 Bytes - hash of a block
        :type block_hash: str
        :return: integer of the number of transactions in this block.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getBlockTransactionCountByHash('0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b')
        69

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getblocktransactioncountbyhash
        .. todo::
            TESTED
        """
        return hex_to_dec(self.call('eth_getBlockTransactionCountByHash', [block_hash]))

    def eth_getBlockTransactionCountByNumber(self, block=BLOCK_TAG_LATEST):
        """ Returns the number of transactions in a block matching the given block number.

        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :return:  integer of the number of transactions in this block.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getBlockTransactionCountByNumber(5100196)
        69

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getblocktransactioncountbynumber
        .. todo::
            TESTED
        """
        block = validate_block(block)
        return hex_to_dec(self.call('eth_getBlockTransactionCountByNumber', [block]))

    def eth_getUncleCountByBlockHash(self, block_hash):
        """ Returns the number of transactions in a block matching the given block number.

        :param block_hash: 32 Bytes - hash of a block
        :type block_hash: str
        :return: integer of the number of uncles in this block.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getUncleCountByBlockHash('0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b')
        0


        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getunclecountbyblockhash
        .. todo::
            TESTED
        """
        return hex_to_dec(self.call('eth_getUncleCountByBlockHash', [block_hash]))

    def eth_getUncleCountByBlockNumber(self, block=BLOCK_TAG_LATEST):
        """ Returns the number of uncles in a block from a block matching the given block number.

        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :return: integer of the number of uncles in this block.
        :rtype: int

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getUncleCountByBlockNumber(5100196)
        0

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getunclecountbyblocknumber
        .. todo::
            TESTED
        """
        block = validate_block(block)
        return hex_to_dec(self.call('eth_getUncleCountByBlockNumber', [block]))

    def eth_getCode(self, address, default_block=BLOCK_TAG_LATEST):
        """ Returns code at a given address.

        :param address: 20 Bytes - address.
        :type address: str
        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :return: the code from the given address.
        :rtype: hex str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getCode("0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413")
        '0x6060604052361561020e5760e060020a6000350463013cf08b[...]62f93160ef3e563'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getcode
        .. todo::
            TESTED
        """
        default_block = validate_block(default_block)
        return self.call('eth_getCode', [address, default_block])

    def eth_sign(self, address, data):
        """ The sign method calculates an Ethereum specific signature with: sign(keccak256("\x19Ethereum Signed Message:\n" + len(message) + message))).

        By adding a prefix to the message makes the calculated signature recognisable as an Ethereum specific signature.
        This prevents misuse where a malicious DApp can sign arbitrary data (e.g. transaction) and use the signature to impersonate the victim.

        :param address: 20 Bytes - address.
        :type address: str
        :param data: N Bytes - message to sign
        :type data: hex str
        :return: Signature
        :rtype: hex str

        .. note::
            the address to sign with must be unlocked.
        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_sign
        .. todo::
            NOT TESTED
        """
        return self.call('eth_sign', [address, data])

    def eth_sendTransaction(self, to_address=None, from_address=None, gas=None, gas_price=None, value=None, data=None,
                            nonce=None):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_sendtransaction

        NEEDS TESTING
        """
        params = {}
        params['from'] = from_address or self.eth_coinbase()
        if to_address is not None:
            params['to'] = to_address
        if gas is not None:
            params['gas'] = hex(gas)
        if gas_price is not None:
            params['gasPrice'] = clean_hex(gas_price)
        if value is not None:
            params['value'] = clean_hex(value)
        if data is not None:
            params['data'] = data
        if nonce is not None:
            params['nonce'] = hex(nonce)
        return self.call('eth_sendTransaction', [params])

    def eth_sendRawTransaction(self, data):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_sendrawtransaction

        NEEDS TESTING
        """
        return self.call('eth_sendRawTransaction', [data])

    def eth_call(self, to_address, from_address=None, gas=None, gas_price=None, value=None, data=None,
                 default_block=BLOCK_TAG_LATEST):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_call

        NEEDS TESTING
        """
        default_block = validate_block(default_block)
        obj = {}
        obj['to'] = to_address
        if from_address is not None:
            obj['from'] = from_address
        if gas is not None:
            obj['gas'] = hex(gas)
        if gas_price is not None:
            obj['gasPrice'] = clean_hex(gas_price)
        if value is not None:
            obj['value'] = value
        if data is not None:
            obj['data'] = data
        return self.call('eth_call', [obj, default_block])

    def eth_estimateGas(self, to_address=None, from_address=None, gas=None, gas_price=None, value=None, data=None,
                        default_block=BLOCK_TAG_LATEST):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_estimategas

        NEEDS TESTING
        """
        if isinstance(default_block, basestring):
            if default_block not in BLOCK_TAGS:
                raise ValueError
        obj = {}
        if to_address is not None:
            obj['to'] = to_address
        if from_address is not None:
            obj['from'] = from_address
        if gas is not None:
            obj['gas'] = hex(gas)
        if gas_price is not None:
            obj['gasPrice'] = clean_hex(gas_price)
        if value is not None:
            obj['value'] = value
        if data is not None:
            obj['data'] = data
        return hex_to_dec(self.call('eth_estimateGas', [obj, default_block]))

    def eth_getBlockByHash(self, block_hash, tx_objects=True):
        """ Returns information about a block by hash.

        :param block_hash: 32 Bytes - Hash of a block.
        :type block_hash: str
        :param tx_objects: (optionnal) If true it returns the full transaction objects, if false only the hashes of the transactions.
        :type tx_objects: Boolean
        :return: A block object, or null when no block was found
        :rtype: dict

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getBlockByHash('0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b')
        {'difficulty': '0xaa41aea7beb9e',
         'extraData': '0x6e616e6f706f6f6c2e6f7267',
         'gasLimit': '0x7a121d',
         'gasUsed': '0x614398',
         'hash': '0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b',
         'logsBloom': '0x000001052440040410040000008006000020000002a11000308045410029410802804801080040c00880000002010c0201804010100900b0000001000240000010800040080044910000010c0000204a041140220008000040000040808800404020802226400018144000400484880012000408000401400000211000c000e2040000209000040080c00000c000890080001090008000001000000102000100002400082240104010400000420080041004050a1080c0042000000000080ac000000802020400001009088021040230000000249208020621000022001820180500200000820002600004888840810420200080100400080000ac0004100000',
         'miner': '0x52bc44d5378309ee2abf1539bf71de1b7d7be3b5',
         'mixHash': '0xa79e0692e7056ea2af26a78a1ed42ac7f3049eb322041c073e5d5a08f6c7e053',
         'nonce': '0x6677371ca8459875',
         'number': '0x4dd2a4',
         [...]
         'totalDifficulty': '0x8b344e12294352eee8',
         'transactions': [{
            [...]
         'value': '0x0'}],
        'transactionsRoot': '0x184cd24c9f45c66ff3846c48fb63e24094aa5909cabfd38211c1d8209128cbc0',
        'uncles': []}

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getblockbyhash
        .. todo::
            TESTED
        """
        return self.call('eth_getBlockByHash', [block_hash, tx_objects])

    def eth_getBlockByNumber(self, block=BLOCK_TAG_LATEST, tx_objects=True):
        """ Returns information about a block by hash.

        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :param tx_objects: (optionnal) If true it returns the full transaction objects, if false only the hashes of the transactions.
        :type tx_objects: Boolean
        :return: A block object, or null when no block was found
        :rtype: dict

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getBlockByNumber(5100196)
        {'difficulty': '0xaa41aea7beb9e',
         'extraData': '0x6e616e6f706f6f6c2e6f7267',
         'gasLimit': '0x7a121d',
         'gasUsed': '0x614398',
         'hash': '0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b',
         'logsBloom': '0x000001052440040410040000008006000020000002a11000308045410029410802804801080040c00880000002010c0201804010100900b0000001000240000010800040080044910000010c0000204a041140220008000040000040808800404020802226400018144000400484880012000408000401400000211000c000e2040000209000040080c00000c000890080001090008000001000000102000100002400082240104010400000420080041004050a1080c0042000000000080ac000000802020400001009088021040230000000249208020621000022001820180500200000820002600004888840810420200080100400080000ac0004100000',
         'miner': '0x52bc44d5378309ee2abf1539bf71de1b7d7be3b5',
         'mixHash': '0xa79e0692e7056ea2af26a78a1ed42ac7f3049eb322041c073e5d5a08f6c7e053',
         'nonce': '0x6677371ca8459875',
         'number': '0x4dd2a4',
         [...]
         'totalDifficulty': '0x8b344e12294352eee8',
         'transactions': [{
            [...]
         'value': '0x0'}],
        'transactionsRoot': '0x184cd24c9f45c66ff3846c48fb63e24094aa5909cabfd38211c1d8209128cbc0',
        'uncles': []}

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getblockbynumber
        .. todo::
            TESTED
        """
        block = validate_block(block)
        return self.call('eth_getBlockByNumber', [block, tx_objects])

    def eth_getTransactionByHash(self, tx_hash):
        """ Returns the information about a transaction requested by transaction hash.

        :param tx_hash:  32 Bytes - hash of a transaction
        :type tx_hash: str
        :return: A transaction object, or null when no transaction was found
        :rtype: dict

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getTransactionByHash('0x345303843c2f3041d12f0c5e6075fd294c2e2ca8cd9b4a9addca3f8caf4380ff')
        {'blockHash': '0xaedd5193cb2b2d9de4c371404277468c7e30eb96f8f9990bd964ca535d88ebc4',
         'blockNumber': '0x4dd3b0',
         'from': '0xc5ff88c3e2902c56c0278fa0e7062d4b5c7e9358',
         'gas': '0x5208',
         'gasPrice': '0xbebc200',
         'hash': '0x345303843c2f3041d12f0c5e6075fd294c2e2ca8cd9b4a9addca3f8caf4380ff',
         'input': '0x',
         'nonce': '0x1',
         'r': '0x5b43f56b69bc78571de39b1b1fc33905a4af588a2ce59f31c54bc54391a255b8',
         's': '0x359e7889447314fb8d9a7e526852471c7bccee006f47222499f9353074107910',
         'to': '0xf954bbabe7bf9a2a4c98dceb293bf437e3863a4e',
         'transactionIndex': '0x66',
         'v': '0x25',
         'value': '0x3202c66793000'}

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gettransactionbyhash
        .. todo::
            TESTED
        """
        return self.call('eth_getTransactionByHash', [tx_hash])

    def eth_getTransactionByBlockHashAndIndex(self, block_hash, index=0):
        """ Returns information about a transaction by block hash and transaction index position.

        :param tx_hash:  32 Bytes - hash of a transaction
        :type tx_hash: str
        :param index: (optionnal) integer of the transaction index position.
        :type index: int
        :return: A transaction object, or null when no transaction was found
        :rtype: dict

        .. seealso::
            :method:`eth_getTransactionByHash`
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gettransactionbyblockhashandindex
        .. todo::
            NOT TESTED
        """
        return self.call('eth_getTransactionByBlockHashAndIndex', [block_hash, hex(index)])

    def eth_getTransactionByBlockNumberAndIndex(self, block=BLOCK_TAG_LATEST, index=0):
        """ Returns information about a transaction by block number and transaction index position.

        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :param index: (optionnal) integer of the transaction index position.
        :type index: int
        :return: A transaction object, or null when no transaction was found
        :rtype: dict

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getTransactionByBlockNumberAndIndex(5100196, 1)
        {'blockHash': '0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b',
         'blockNumber': '0x4dd2a4',
         'from': '0xb01cb49fe0d6d6e47edf3a072d15dfe73155331c',
         'gas': '0x5208',
         'gasPrice': '0xe33e22200',
         'hash': '0xf02ffa405bae96e62a9e36fbd781362ca378ec62353d5e2bd0585868d3deaf61',
         'input': '0x',
         'nonce': '0x1908f',
         'r': '0xcad900a5060ba9bb646a7f6965f98e945d71a19b3e30ff53d03b9797c6153d07',
         's': '0x53b11a48758fc383df878a9b5468c83b033f5036b124b16dbb0a5167aee7fc4f',
         'to': '0x26cd018553871f2e887986bc24c68a0ce622bb8f',
         'transactionIndex': '0x1',
         'v': '0x25',
         'value': '0x1bc16d674ec80000'}

        .. seealso::
            :method:`eth_getTransactionByHash`
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gettransactionbyblocknumberandindex
        .. todo::
            TESTED
        """
        block = validate_block(block)
        return self.call('eth_getTransactionByBlockNumberAndIndex', [block, hex(index)])

    def eth_getTransactionReceipt(self, tx_hash):
        """ Returns the receipt of a transaction by transaction hash.

        :param tx_hash: 32 Bytes - hash of a transaction
        :type tx_hash: str
        :return: A transaction receipt object, or null when no receipt was found
        :rtype: dict

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_getTransactionReceipt('0xf02ffa405bae96e62a9e36fbd781362ca378ec62353d5e2bd0585868d3deaf61')
        {'blockHash': '0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b',
         'blockNumber': '0x4dd2a4',
         'contractAddress': None,
         'cumulativeGasUsed': '0xe2d7',
         'from': '0xb01cb49fe0d6d6e47edf3a072d15dfe73155331c',
         'gasUsed': '0x5208',
         'logs': [],
         'logsBloom': '0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000',
         'status': '0x1',
         'to': '0x26cd018553871f2e887986bc24c68a0ce622bb8f',
         'transactionHash': '0xf02ffa405bae96e62a9e36fbd781362ca378ec62353d5e2bd0585868d3deaf61',
         'transactionIndex': '0x1'}

        .. note::
            That the receipt is not available for pending transactions.
        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_gettransactionreceipt
        .. todo::
            TESTED

        """
        return self.call('eth_getTransactionReceipt', [tx_hash])

    def eth_getUncleByBlockHashAndIndex(self, block_hash, index=0):
        """ Returns information about a uncle of a block by hash and uncle index position.

        :param tx_hash:  32 Bytes - hash of a transaction
        :type tx_hash: str
        :param index: (optionnal)  the uncle's index position.
        :type index: int
        :return: A block object, or null when no block was found
        :rtype: dict

        .. note::
            An uncle doesn't contain individual transactions.
        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getunclebyblockhashandindex
        .. todo::
            NOT TESTED
        """
        return self.call('eth_getUncleByBlockHashAndIndex', [block_hash, hex(index)])

    def eth_getUncleByBlockNumberAndIndex(self, block=BLOCK_TAG_LATEST, index=0):
        """ Returns information about a uncle of a block by number and uncle index position.

        :param block: (optionnal) integer block number, or the string "latest", "earliest" or "pending"
        :type block: int or str
        :param index: (optionnal) the uncle's index position.
        :type index: int
        :return: A block object, or null when no block was found
        :rtype: dict

        .. note::
            An uncle doesn't contain individual transactions.
        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getunclebyblocknumberandindex
        .. todo::
            NOT TESTED
        """
        block = validate_block(block)
        return self.call('eth_getUncleByBlockNumberAndIndex', [block, hex(index)])

    def eth_getCompilers(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getcompilers

        NOT WORKING
        """
        return self.call('eth_getCompilers')

    def eth_compileSolidity(self, code):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_compilesolidity

        NOT WORKING
        """
        return self.call('eth_compileSolidity', [code])

    def eth_compileLLL(self, code):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_compilelll

        N/A
        """
        return self.call('eth_compileLLL', [code])

    def eth_compileSerpent(self, code):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_compileserpent

        N/A
        """
        return self.call('eth_compileSerpent', [code])

    def eth_newFilter(self, from_block=BLOCK_TAG_LATEST, to_block=BLOCK_TAG_LATEST, address=None, topics=None):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_newfilter

        NEEDS TESTING
        """
        _filter = {
            'fromBlock': from_block,
            'toBlock': to_block,
            'address': address,
            'topics': topics,
        }
        return self.call('eth_newFilter', [_filter])

    def eth_newBlockFilter(self):
        """ Creates a filter in the node, to notify when a new block arrives. To check if the state has changed, call eth_getFilterChanges.

        :return: A filter id.
        :rtype: hex str

        :Example:

        >>> explorer = EthereumExplorerRPC()
        >>> explorer.eth_newBlockFilter()
        '0x1d21d3c44b9a1501d4358a44bdb6da1d'

        .. seealso::
            https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_newblockfilter
        .. todo::
            TESTED
        """
        return self.call('eth_newBlockFilter')

    def eth_newPendingTransactionFilter(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_newpendingtransactionfilter

        TESTED
        """
        return hex_to_dec(self.call('eth_newPendingTransactionFilter'))

    def eth_uninstallFilter(self, filter_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_uninstallfilter

        NEEDS TESTING
        """
        return self.call('eth_uninstallFilter', [filter_id])

    def eth_getFilterChanges(self, filter_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getfilterchanges

        NEEDS TESTING
        """
        return self.call('eth_getFilterChanges', [filter_id])

    def eth_getFilterLogs(self, filter_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getfilterlogs

        NEEDS TESTING
        """
        return self.call('eth_getFilterLogs', [filter_id])

    def eth_getLogs(self, filter_object):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getlogs

        NEEDS TESTING
        """
        return self.call('eth_getLogs', [filter_object])

    def eth_getWork(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_getwork

        TESTED
        """
        return self.call('eth_getWork')

    def eth_submitWork(self, nonce, header, mix_digest):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_submitwork

        NEEDS TESTING
        """
        return self.call('eth_submitWork', [nonce, header, mix_digest])

    def eth_submitHashrate(self, hash_rate, client_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_submithashrate

        TESTED
        """
        return self.call('eth_submitHashrate', [hex(hash_rate), client_id])

    def db_putString(self, db_name, key, value):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#db_putstring

        TESTED
        """
        warnings.warn('deprecated', DeprecationWarning)
        return self.call('db_putString', [db_name, key, value])

    def db_getString(self, db_name, key):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#db_getstring

        TESTED
        """
        warnings.warn('deprecated', DeprecationWarning)
        return self.call('db_getString', [db_name, key])

    def db_putHex(self, db_name, key, value):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#db_puthex

        TESTED
        """
        if not value.startswith('0x'):
            value = '0x{}'.format(value)
        warnings.warn('deprecated', DeprecationWarning)
        return self.call('db_putHex', [db_name, key, value])

    def db_getHex(self, db_name, key):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#db_gethex

        TESTED
        """
        warnings.warn('deprecated', DeprecationWarning)
        return self.call('db_getHex', [db_name, key])

    def shh_version(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_version

        N/A
        """
        return self.call('shh_version')

    def shh_post(self, topics, payload, priority, ttl, from_=None, to=None):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_post

        NEEDS TESTING
        """
        whisper_object = {
            'from': from_,
            'to': to,
            'topics': topics,
            'payload': payload,
            'priority': hex(priority),
            'ttl': hex(ttl),
        }
        return self.call('shh_post', [whisper_object])

    def shh_newIdentity(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_newidentity

        N/A
        """
        return self.call('shh_newIdentity')

    def shh_hasIdentity(self, address):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_hasidentity

        NEEDS TESTING
        """
        return self.call('shh_hasIdentity', [address])

    def shh_newGroup(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_newgroup

        N/A
        """
        return self.call('shh_newGroup')

    def shh_addToGroup(self):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_addtogroup

        NEEDS TESTING
        """
        return self.call('shh_addToGroup')

    def shh_newFilter(self, to, topics):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_newfilter

        NEEDS TESTING
        """
        _filter = {
            'to': to,
            'topics': topics,
        }
        return self.call('shh_newFilter', [_filter])

    def shh_uninstallFilter(self, filter_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_uninstallfilter

        NEEDS TESTING
        """
        return self.call('shh_uninstallFilter', [filter_id])

    def shh_getFilterChanges(self, filter_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_getfilterchanges

        NEEDS TESTING
        """
        return self.call('shh_getFilterChanges', [filter_id])

    def shh_getMessages(self, filter_id):
        """
        https://github.com/ethereum/wiki/wiki/JSON-RPC#shh_getmessages

        NEEDS TESTING
        """
        return self.call('shh_getMessages', [filter_id])


class EthereumInfuraExplorer(EthereumExplorerRPC):
    """
    EthereumExplorer subclass for using Infura
    """

    def __init__(self, key, network=INFURA_MAINNET):
        EthereumExplorerRPC.__init__(self, host=network + key,
                                     port=INFURA_RPC_PORT, tls=True)


class EthereumParityExplorer(EthereumExplorerRPC):
    """
    EthereumExplorer subclass for Parity-specific methods
    """

    def __init__(self, host='localhost', port=PARITY_DEFAULT_RPC_PORT, tls=False):
        EthereumExplorerRPC.__init__(self, host=host, port=port, tls=tls)

    def trace_filter(self, from_block=None, to_block=None, from_addresses=None, to_addresses=None):
        """
        https://github.com/ethcore/parity/wiki/JSONRPC-trace-module#trace_filter

        TESTED
        """
        params = {}
        if from_block is not None:
            from_block = validate_block(from_block)
            params['fromBlock'] = from_block
        if to_block is not None:
            to_block = validate_block(to_block)
            params['toBlock'] = to_block
        if from_addresses is not None:
            if not isinstance(from_addresses, list):
                from_addresses = [from_addresses]
            params['fromAddress'] = from_addresses
        if to_addresses is not None:
            if not isinstance(to_addresses, list):
                to_addresses = [to_addresses]
            params['toAddress'] = to_addresses
        return self.call('trace_filter', [params])

    def trace_get(self, tx_hash, positions):
        """
        https://wiki.parity.io/JSONRPC
        https://github.com/ethcore/parity/wiki/JSONRPC-trace-module#trace_get

        NEEDS TESTING
        """
        if not isinstance(positions, list):
            positions = [positions]
        return self.call('trace_get', [tx_hash, positions])

    def trace_transaction(self, tx_hash):
        """
        https://wiki.parity.io/JSONRPC
        https://github.com/ethcore/parity/wiki/JSONRPC-trace-module#trace_transaction

        NEEDS TESTING
        """
        return self.call('trace_transaction', [tx_hash])

    def trace_block(self, block=BLOCK_TAG_LATEST):
        """
        https://wiki.parity.io/JSONRPC
        https://github.com/ethcore/parity/wiki/JSONRPC-trace-module#trace_block

        NEEDS TESTING
        """
        block = validate_block(block)
        return self.call('trace_block', [block])



import io

from octopus.engine.disassembler import Disassembler

from octopus.platforms.NEO.instruction import NeoInstruction
from octopus.platforms.NEO.avm import Avm


class NeoDisassembler(Disassembler):

    def __init__(self, bytecode=None):
        Disassembler.__init__(self, bytecode=bytecode, asm=Avm())

    def disassemble_opcode(self, bytecode, offset=0):
        '''
        TODO
        '''

        wallet = io.BytesIO(bytecode)
        opcode = int.from_bytes(wallet.read(1), byteorder='big')

        invalid = ('INVALID', 0, 0, 0, 0, 'Unknown opcode')
        name, operand_size, pops, pushes, gas, description = \
            self.asm.table.get(opcode, invalid)
        instruction = NeoInstruction(opcode, name, operand_size, pops, pushes,
                                     gas, description, offset=offset)
        if instruction.has_length_operand:
            instruction.operand_size = int.from_bytes(wallet.read(1), byteorder='big')
        if instruction.has_operand:
            instruction.operand = wallet.read(instruction.operand_size)
        return instruction

    def disassemble(self, bytecode=None, offset=0, r_format='list'):
        '''
        TODO
        '''
        self.bytecode = str(bytecode) if bytecode else str(self.bytecode)

        self.instructions = list()
        self.reverse_instructions = dict()

        return super().disassemble(bytecode, offset, r_format)



from graphviz import Digraph

from octopus.core.edge \
    import (EDGE_UNCONDITIONAL,
            EDGE_CONDITIONAL_TRUE, EDGE_CONDITIONAL_FALSE,
            EDGE_FALLTHROUGH, EDGE_CALL)

import logging
log = logging.getLogger(__name__)
log.setLevel(level=logging.DEBUG)


def insert_edges_to_graph(graph, edges, call=False):
    # remove duplicate edges
    edges = list(set(edges))

    # create link between block
    for edge in edges:

        if edge.type == EDGE_UNCONDITIONAL:
            graph.edge(edge.node_from, edge.node_to, color='blue')
        elif edge.type == EDGE_CONDITIONAL_TRUE:
            graph.edge(edge.node_from, edge.node_to, color='green')
        elif edge.type == EDGE_CONDITIONAL_FALSE:
            graph.edge(edge.node_from, edge.node_to, color='red')
        elif edge.type == EDGE_FALLTHROUGH:
            graph.edge(edge.node_from, edge.node_to, color='cyan')
        elif edge.type == EDGE_CALL and call:
            graph.edge(edge.node_from, edge.node_to, color='yellow')
        else:
            raise Exception('Edge type unknown')


class Graph(object):

    def __init__(self, basicblocks, edges, functions=None,
                 filename='graph.gv', design=None):
        self.basicblocks = basicblocks
        self.edges = edges
        self.filename = filename
        self.design = design or {'shape': 'box', 'fontname': 'Courier',
                                 'fontsize': '30.0', 'rank': 'same'}

    def view_ssa(self, call=False, view=True):
        self.view(view=view, call=call, ssa=True)

    def view_simplify(self, call=False, view=True):
        self.view(view=view, call=call, simplify=True)

    def view(self, view=True, simplify=False, call=False, ssa=False):
        g = Digraph(self.filename, filename=self.filename)

        with g.subgraph(name='global', node_attr=self.design) as c:
            c.label = 'global'

            # create all the basicblocks (blocks)
            for basicblock in self.basicblocks:
                if simplify:
                    # create node
                    c.node(basicblock.name, label=basicblock.name)
                else:
                    if ssa:
                        label = basicblock.instructions_ssa()
                    else:
                        label = basicblock.instructions_details()
                    # the escape sequences "\n", "\l" and "\r"
                    # divide the label into lines, centered,
                    # left-justified, and right-justified, respectively.
                    label = label.replace('\n', '\l')
                    # create node
                    c.node(basicblock.name, label=label)

        # insert edges on the graph
        insert_edges_to_graph(g, self.edges, call)

        g.render(self.filename, view=view)
        # g.view()


class CallGraph(object):
    def __init__(self, nodes, edges,
                 filename='call_graph.gv', design=None):
        self.nodes = nodes
        self.edges = edges
        self.filename = filename
        self.design = design or {'shape': 'box'}

    def view(self, view=True):
        g = Digraph(self.filename, filename=self.filename,
                    graph_attr=self.design)
        g.attr(rankdir='LR')

        with g.subgraph(name='global', node_attr=self.design) as c:

            # create all the graph nodes (function name)
            for idx, node in enumerate(self.nodes):
                # print('%d %s' % (idx, node))
                c.node(node)

            # check if multiple same edges
            # in that case, put the number into label
            edges_counter = dict((x,self.edges.count(x)) for x in set(self.edges))
            # insert edges on the graph
            for edge, count in edges_counter.items():
                label = None
                if count > 1:
                    label = str(count)
                c.edge(edge.node_from, edge.node_to, label=label, color='black')

        g.render(self.filename, view=view)


class CFGGraph(Graph):

    def __init__(self, cfg, filename='graph.cfg.gv', design=None):
        Graph.__init__(self, cfg.basicblocks, cfg.edges, filename=filename,
                       design=design)
        self.cfg = cfg

    def view_functions_ssa(self, call=False, view=True):
        self.view_functions(view=view, call=call, ssa=True)

    def view_functions_simplify(self, call=False, view=True):
        self.view_functions(view=view, call=call, simplify=True)

    def view_functions(self, only_func_name=None, view=True, simplify=False, call=False, ssa=False, color='grey'):
        g = Digraph('G', filename=self.filename)
        g.attr(rankdir='TB')
        g.attr(overlap='scale')
        g.attr(splines='polyline')
        g.attr(ratio='fill')

        functions = self.cfg.functions
        # only show functions listed
        if only_func_name:
            functions = [func for func in self.cfg.functions if func.name in only_func_name]

        for idx, func in enumerate(functions):
            with g.subgraph(name='cluster_%d' % idx, node_attr=self.design) as c:
                if func.name == func.prefered_name:
                    name = func.name
                else:
                    name = func.prefered_name + ' - ' + func.name
                c.attr(label=name)
                c.attr(color=color)
                c.attr(fontsize='50.0')
                c.attr(overlap='false')
                c.attr(splines='polyline')
                c.attr(ratio='fill')

                # create all the basicblocks (blocks)
                for basicblock in func.basicblocks:
                    if simplify:
                        # create node
                        c.node(basicblock.name, label=basicblock.name, splines='true')
                    else:
                        if ssa:
                            label = basicblock.instructions_ssa()
                        else:
                            label = basicblock.instructions_details()
                        # the escape sequences "\n", "\l" and "\r"
                        # divide the label into lines, centered,
                        # left-justified, and right-justified, respectively.
                        label = label.replace('\n', '\l')
                        # create node
                        c.node(basicblock.name, label=label)

        edges = self.cfg.edges
        # only get corresponding edges
        if only_func_name:
            functions_block = [func.basicblocks for func in functions]
            block_name = [b.name for block_l in functions_block for b in block_l]
            edges = [edge for edge in edges if (edge.node_from in block_name or edge.node_to in block_name)]
        # insert edges on the graph
        insert_edges_to_graph(g, edges, call)

        g.render(self.filename, view=view)
        # g.view()



from octopus.engine.explorer import Explorer
from octopus.config.EOS_EXPLORER import EOS_DEFAULT_RPC_PORT

from requests.exceptions import ConnectionError as RequestsConnectionError
import json

EOS_DEFAULT_RPC_PORT = 8888
EOS_WALLET_RPC_PORT = 8889

class EosExplorer(Explorer):
    """
    EOS REST RPC client class


    doc: https://eosio.github.io/eos/group__eosiorpc.html
    cleos source code: https://github.com/EOSIO/eos/blob/master/programs/cleos/main.cpp
    """
    def __init__(self, host='localhost', port=EOS_DEFAULT_RPC_PORT, tls=False, max_retries=3):
        Explorer.__init__(self, host=host, port=port, tls=tls, max_retries=max_retries)

    def call(self, method, params={}, version='v1', api_type='chain'):

        current_url = '{}/{}/{}/{}'.format(self.url, version, api_type, method)

        try:
            r = self.session.post(current_url, headers=self.headers, data=json.dumps(params))
        except RequestsConnectionError:
            raise Exception('RPC connection Error')
        if not 200 <= r.status_code < 300:
            raise Exception('RPC connection failure: ' + str(r.status_code) + ' ' + r.reason + ' ' + r.text)
        try:
            response = r.json()
        except ValueError:
            raise Exception('JSON response parsing error: ' + str(r.text))
        try:
            return response
        except KeyError:
            raise Exception('\"result\" field in JSON response error: ' + str(response))

    ##########################
    #        Chain API       #
    ##########################

    def get_info(self):
        '''Get latest information related to a node

        TESTED
        '''
        return self.call('get_info')

    def get_block(self, block_num_or_id):
        '''Get information related to a block.

        TESTED
        '''
        data = {'block_num_or_id': block_num_or_id}
        return self.call('get_block', data)

    def get_account(self, account_name):
        '''Get information related to an account.

        TESTED
        '''
        data = {'account_name': account_name}
        return self.call('get_account', data)

    def get_code(self, account_name):
        '''Fetch smart contract code.

        TESTED
        '''
        data = {'account_name': account_name}
        return self.call('get_code', data)

    def get_table_rows(self, scope, code, table, json=False, lower_bound=None, upper_bound=None, limit=None):
        '''Fetch smart contract data from an account.

        NOT TESTED
        '''
        data = {'scope': scope,
                'code': code,
                'table': table,
                'json': json}
        if lower_bound:
            data['lower_bound'] = lower_bound
        if upper_bound:
            data['upper_bound'] = upper_bound
        if limit:
            data['limit'] = limit
        return self.call('get_table_rows', data)

    def abi_json_to_bin(self, code, action, args):
        '''Serialize json to binary hex. The resulting binary hex is usually used for the data field in push_transaction.

        NOT TESTED
        '''
        data = {'code': code,
                'action': action,
                'args': args}
        print(data)
        return self.call('abi_json_to_bin', data)

    def abi_bin_to_json(self, code, action, binargs):
        '''Serialize back binary hex to json.

        NOT TESTED
        '''
        data = {'code': code,
                'action': action,
                'binargs': binargs}
        return self.call('abi_bin_to_json', data)

    def push_transaction(self, tx_json):
        '''This method expects a transaction in JSON format and will attempt to apply it to the blockchain,

        NOT TESTED
        '''
        return self.call('push_transaction', tx_json)

    def push_transactions(self, list_tx_json):
        '''This method push multiple transactions at once.

        NOT TESTED
        '''
        return self.call('push_transactions', list_tx_json)

    def get_required_keys(self, transaction):
        '''Get required keys to sign a transaction from list of your keys.

        NOT TESTED
        '''
        data = {'transaction': transaction}
        return self.call('get_required_keys', data)

    ##########################
    #       Wallet API       #
    ##########################

    def wallet_create(self, name):
        '''Create a new wallet with the given name

        TESTED
        '''
        return self.call('create', name, api_type='wallet')

    def wallet_open(self, name):
        '''Open an existing wallet of the given name

        TESTED
        '''
        return self.call('open', name, api_type='wallet')

    def wallet_lock(self, name):
        '''Lock a wallet of the given name

        TESTED
        '''
        return self.call('lock', name, api_type='wallet')

    def wallet_lock_all(self):
        '''Lock all wallets

        TESTED
        '''
        return self.call('lock_all', api_type='wallet')

    def wallet_unlock(self, name, password):
        '''Unlock a wallet with the given name and password

        TESTED
        '''
        return self.call('unlock', [name, password], api_type='wallet')

    def wallet_import_key(self, name, priv_key):
        '''Import a private key to the wallet of the given name

        TESTED
        '''
        return self.call('import_key', [name, priv_key], api_type='wallet')

    def wallet_list(self):
        '''List all wallets

        TESTED
        '''
        return self.call('list_wallets', api_type='wallet')

    def wallet_list_keys(self):
        '''List all key pairs across all wallets

        TESTED
        '''
        return self.call('list_keys', api_type='wallet')

    def wallet_get_public_keys(self):
        '''List all public keys across all wallets

        TESTED
        '''
        return self.call('get_public_keys', api_type='wallet')

    def wallet_set_timeout(self, timeout_s):
        '''Set wallet auto lock timeout (in seconds)

        TESTED
        '''
        return self.call('set_timeout', timeout_s, api_type='wallet')

    def wallet_sign_trx(self, tx_json):
        '''Sign transaction given an array of transaction, require public keys, and chain id

        NOT TESTED
        '''
        return self.call('sign_transaction', tx_json)



import io
import logging
import re

from octopus.engine.disassembler import Disassembler

from octopus.arch.evm.instruction import EvmInstruction
from octopus.arch.evm.evm import EVM


def runtime_code_detector(bytecode_hex):
    '''Check for presence of runtime code
    '''
    result = list(re.finditer('60.{2}604052', bytecode_hex))
    if len(result) > 1:
        position = result[1].start()
        logging.info("[+] Runtime code detected")
        return bytecode_hex[position:]
    return bytecode_hex


def swarm_hash_detector(bytecode_hex):
    '''Check for presence of Swarm hash at the end of bytecode
        https://github.com/ethereum/wiki/wiki/Swarm-Hash
    '''
    swarm_hash = bytecode_hex[-86:]
    # bzzr == 0x627a7a72
    if '627a7a72' in swarm_hash:
        logging.info("[+] Swarm hash detected in bytecodes")
        bytecode_hex = bytecode_hex[:-86]
        logging.info("[+] Swarm hash value: 0x%s", swarm_hash)
        logging.info("[+] Swarm hash removed")
    return bytecode_hex


class EvmDisassembler(Disassembler):

    def __init__(self, bytecode=None):
        Disassembler.__init__(self, asm=EVM(), bytecode=bytecode)

    def disassemble_opcode(self, bytecode, offset=0):
        """
        TODO
        """

        wallet = io.BytesIO(bytecode)
        opcode = int.from_bytes(wallet.read(1), byteorder='big')

        # default value
        invalid = ('INVALID', 0, 0, 0, 0, 'Unknown opcode')
        name, operand_size, pops, pushes, gas, description = \
            self.asm.table.get(opcode, invalid)
        instruction = EvmInstruction(opcode, name, operand_size, pops, pushes,
                                     gas, description, offset=offset)
        if instruction.has_operand:
            instruction.operand = wallet.read(operand_size)
            if instruction.is_push:
                # directly calculate the operand int representation
                instruction.operand_interpretation = \
                    int.from_bytes(instruction.operand, byteorder='big')

        return instruction

    def disassemble(self, bytecode=None, offset=0, r_format='list',
                    analysis=True):
        '''
        TODO
        '''

        self.bytecode = str(bytecode) if bytecode else str(self.bytecode)

        if analysis:
            self.bytecode = runtime_code_detector(self.bytecode)
            self.bytecode = swarm_hash_detector(self.bytecode)

        self.instructions = list()
        self.reverse_instructions = dict()

        return super().disassemble(self.bytecode, offset,
                                   r_format)



import unittest

from octopus.platforms.NEO.disassembler import NeoDisassembler


class NeoDisassemblerTestCase(unittest.TestCase):

    def testDisassemble(self):

        def disassemble(bytecode_hex, result):
            disasm = len(NeoDisassembler(bytecode_hex).disassemble())
            self.assertEqual(disasm, result)

        def disasmOne(bytecode, result):
            disasm = str(NeoDisassembler(bytecode).disassemble_opcode(bytecode))
            self.assertEqual(disasm, result)

        def disasm(bytecode, result):
            disasm = NeoDisassembler(bytecode_hex).disassemble(r_format='text')
            self.assertEqual(disasm, result)

        bytecode = b'Q\xc5kaah\x16Neo.Storage.GetContext\x06secret\x0bHello WorldaRrh\x0fNeo.Storage.Putaah\x16Neo.Storage.GetContext\x06secreta|h\x0fNeo.Storage.Getlvk\x00Rz\xc4b\x03\x00lvk\x00\xc3aluf'
        bytecode_hex = "0x51c56b616168164e656f2e53746f726167652e476574436f6e74657874067365637265740b48656c6c6f20576f726c64615272680f4e656f2e53746f726167652e507574616168164e656f2e53746f726167652e476574436f6e7465787406736563726574617c680f4e656f2e53746f726167652e4765746c766b00527ac46203006c766b00c3616c7566"
        result = 'PUSH1\nNEWARRAY\nTOALTSTACK\nNOP\nNOP\nSYSCALL 0x4e656f2e53746f726167652e476574436f6e74657874\nPUSHBYTES6 0x736563726574\nPUSHBYTES11 0x48656c6c6f20576f726c64\nNOP\nPUSH2\nXSWAP\nSYSCALL 0x4e656f2e53746f726167652e507574\nNOP\nNOP\nSYSCALL 0x4e656f2e53746f726167652e476574436f6e74657874\nPUSHBYTES6 0x736563726574\nNOP\nSWAP\nSYSCALL 0x4e656f2e53746f726167652e476574\nFROMALTSTACK\nDUP\nTOALTSTACK\nPUSH0\nPUSH2\nROLL\nSETITEM\nJMP 0x300\nFROMALTSTACK\nDUP\nTOALTSTACK\nPUSH0\nPICKITEM\nNOP\nFROMALTSTACK\nDROP\nRET'

        bytecode_one_instr = b'h\x16Neo.Storage.GetContext'
        result_one_instr = 'SYSCALL 0x4e656f2e53746f726167652e476574436f6e74657874'

        disasmOne(bytecode_one_instr, result_one_instr)
        disasm(bytecode, result)
        disassemble(bytecode_hex, 36)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(NeoDisassembler)
    unittest.TextTestRunner(verbosity=2).run(suite)



class Contract(object):

    def __init__(self, name="default_contract_name",
                 address=None, balance=None, bytecode=None,
                 abi=None, source_code=None):
        self.name = name
        self.address = address
        self.balance = balance
        self.bytecode = bytecode
        self.abi = abi
        self.source_code = source_code
        self.info = {}

    def get_online_bytecode(self, explorer):
        raise NotImplementedError

    def get_online_abi(self, explorer):
        raise NotImplementedError

    def get_online_source(self, explorer):
        raise NotImplementedError

    def get_online_info(self, explorer):
        raise NotImplementedError



from octopus.core.function import Function
from octopus.core.basicblock import BasicBlock
from octopus.core.edge import (Edge,
                               EDGE_UNCONDITIONAL,
                               EDGE_CONDITIONAL_TRUE, EDGE_CONDITIONAL_FALSE,
                               EDGE_FALLTHROUGH, EDGE_CALL)
from octopus.analysis.cfg import CFG

from octopus.arch.wasm.analyzer import WasmModuleAnalyzer
from octopus.arch.wasm.disassembler import WasmDisassembler
from octopus.arch.wasm.format import (format_func_name,
                                      format_bb_name)


from octopus.core.utils import bytecode_to_bytes
import logging

# for graph visualisation
from graphviz import Digraph
DESIGN_IMPORT = {'fillcolor': 'turquoise',
                 'shape': 'box',
                 'style': 'filled'}

DESIGN_EXPORT = {'fillcolor': 'grey',
                 'shape': 'box',
                 'style': 'filled'}

log = logging.getLogger(__name__)
log.setLevel(level=logging.WARNING)


def enum_func(module_bytecode):
    ''' return a list of Function
        see:: octopus.core.function
    '''
    functions = list()
    analyzer = WasmModuleAnalyzer(module_bytecode)

    protos = analyzer.func_prototypes
    import_len = len(analyzer.imports_func)

    for idx, code in enumerate(analyzer.codes):
        # get corresponding function prototype
        name, param_str, return_str = protos[import_len + idx]

        name = format_func_name(name, param_str, return_str)
        instructions = WasmDisassembler().disassemble(code)
        cur_function = Function(0, instructions[0], name=name)
        cur_function.instructions = instructions

        functions.append(cur_function)
    return functions


def enum_func_call_edges(functions, len_imports):
    ''' return a list of tuple with
        (index_func_node_from, index_func_node_to)
    '''
    call_edges = list()

    # iterate over functions
    for index, func in enumerate(functions):
        node_from = len_imports + index
        # iterates over instruction
        for inst in func.instructions:
            # detect if inst is a call instructions
            if inst.is_call:
                log.info('%s', inst.operand_interpretation)
                if inst.name == "call":
                    # only get the import_id
                    node_to = int(inst.operand_interpretation.split(' ')[1])
                # The `call_indirect` operator takes a list of function arguments and as the last operand the index into the table.
                elif inst.name == "call_indirect":
                    # the last operand is the index on the table
                    node_to = int(inst.operand_interpretation.split(',')[-1].split(' ')[-1])
                call_edges.append((node_from, node_to))

    return call_edges


def enum_blocks_edges(function_id, instructions):

    """
    Return a list of basicblock after
    statically parsing given instructions
    """

    basicblocks = list()
    edges = list()

    branches = []
    xrefs = []

    intent = 0
    blocks_tmp = []
    blocks_list = []
    # remove last instruction that is 'end' for the funtion
    #tt = instructions[:-1]
    for index, inst in enumerate(instructions[:-1]):

        if inst.is_block_terminator:
            start, name = blocks_tmp.pop()
            if inst.name == 'else':
                end = inst.offset - 1
            else:
                end = inst.offset_end
            blocks_list.append((intent, start, end, name))
            intent -= 1
        if inst.is_block_starter:  # in ['block', 'loop']:
            blocks_tmp.append((inst.offset, inst.name))
            intent += 1
        if inst.is_branch:
            branches.append((intent, inst))

    # add function body end
    blocks_list.append((0, 0, instructions[-1].offset_end, 'func'))
    blocks_list = sorted(blocks_list, key=lambda tup: (tup[1], tup[0]))

    for depth, inst in branches:
        labl = list()
        if inst.name == 'br_table':
            labl = [i for i in inst.insn_byte[2:]]
        else:
            labl.append(int(inst.operand_interpretation.split(' ')[-1]))

        for d2 in labl:
            rep = next(((i, s, e, n) for i, s, e, n in blocks_list if (i == (depth - d2) and s < inst.offset and e > inst.offset_end)), None)
            if rep:
                i, start, end, name = rep
                if name == 'loop':
                    value = start  # else name == 'block'
                elif name == 'block' or name == 'func':
                    value = end
                else:
                    value = None
                inst.xref.append(value)
                xrefs.append(value)

    # remove "block" instruction - not usefull graphicaly
    # instructions = [x for x in instructions if x.name not in ['block', 'loop']]

    # enumerate blocks

    new_block = True

    for index, inst in enumerate(instructions):

        # creation of a block
        if new_block:
            block = BasicBlock(inst.offset,
                               inst,
                               name=format_bb_name(function_id, inst.offset))
            new_block = False

        # add current instruction to the basicblock
        block.instructions.append(inst)

        # next instruction is a jump target
        if index < (len(instructions) - 1) and \
           instructions[index + 1].offset in xrefs:
            new_block = True
        # absolute jump - br
        elif inst.is_branch_unconditional:
            new_block = True
        # conditionnal jump - br_if
        elif inst.is_branch_conditional:
            new_block = True
        # end of a block
        elif index < (len(instructions) - 1) and \
                inst.name in ['end'] and block.start_offset != inst.offset:  # is_block_terminator
            new_block = True
        #elif index < (len(instructions) - 1) and \
        #        instructions[index + 1].name == 'else':  # is_block_terminator
        #    new_block = True
        # start of a block
        elif index < (len(instructions) - 1) and \
                instructions[index + 1].is_block_starter:
            new_block = True
        # last instruction of the bytecode
        elif inst.offset == instructions[-1].offset:
            new_block = True

        if new_block:
            block.end_offset = inst.offset_end
            block.end_instr = inst
            basicblocks.append(block)
            new_block = True

    # TODO: detect and remove end instruction that end loop

    # enumerate edges
    for index, block in enumerate(basicblocks):
        # get the last instruction
        inst = block.end_instr
        # unconditional jump - br
        if inst.is_branch_unconditional:
            for ref in inst.xref:
                edges.append(Edge(block.name, format_bb_name(function_id, ref), EDGE_UNCONDITIONAL))
        # conditionnal jump - br_if, if
        elif inst.is_branch_conditional:
            if inst.name == 'if':
                edges.append(Edge(block.name,
                             format_bb_name(function_id, inst.offset_end + 1),
                             EDGE_CONDITIONAL_TRUE))
                g_block = next(iter([b for b in blocks_list if b[1] == inst.offset]), None)
                jump_target = g_block[2] + 1
                edges.append(Edge(block.name,
                             format_bb_name(function_id, jump_target),
                             EDGE_CONDITIONAL_FALSE))
            else:
                for ref in inst.xref:
                    if ref != inst.offset_end + 1:
                        edges.append(Edge(block.name,
                                          format_bb_name(function_id, ref),
                                          EDGE_CONDITIONAL_TRUE))
                edges.append(Edge(block.name,
                             format_bb_name(function_id, inst.offset_end + 1),
                             EDGE_CONDITIONAL_FALSE))
        elif [i.name for i in block.instructions if i.is_halt]:
            pass
        elif inst.is_halt:
            pass
        elif inst.offset != instructions[-1].offset:
            # EDGE_FALLTHROUGH
            edges.append(Edge(block.name, format_bb_name(function_id, inst.offset_end + 1), EDGE_FALLTHROUGH))

    # prevent duplicate edges
    edges = list(set(edges))
    return basicblocks, edges


class WasmCFG(CFG):
    """
    """
    def __init__(self, module_bytecode, static_analysis=True):

        self.module_bytecode = bytecode_to_bytes(module_bytecode)
        self.static_analysis = static_analysis
        self.analyzer = None

        self.functions = list()
        self.basicblocks = list()
        self.edges = list()

        if self.static_analysis:
            self.analyzer = WasmModuleAnalyzer(self.module_bytecode)
            self.run_static_analysis()

    def run_static_analysis(self):
        self.functions = enum_func(self.module_bytecode)

        for idx, func in enumerate(self.functions):
            func.basicblocks, edges = enum_blocks_edges(idx, func.instructions)
            # all bb name are unique so we can create global bb & edge list
            self.basicblocks += func.basicblocks
            self.edges += edges

    def get_functions_call_edges(self, format_fname=True):

        nodes = list()
        edges = list()

        if not self.analyzer:
            self.analyzer = WasmModuleAnalyzer(self.module_bytecode)
        if not self.functions:
            self.functions = enum_func(self.module_bytecode)

        # create nodes
        for name, param_str, return_str in self.analyzer.func_prototypes:
            if format_fname:
                nodes.append(format_func_name(name, param_str, return_str))
            else:
                nodes.append(name)

        log.info('nodes: %s', nodes)

        # create edges
        tmp_edges = enum_func_call_edges(self.functions,
                                         len(self.analyzer.imports_func))

        # tmp_edges = [(node_from, node_to), (...), ...]
        for node_from, node_to in tmp_edges:
            # node_from
            name, param, ret = self.analyzer.func_prototypes[node_from]
            if format_fname:
                from_final = format_func_name(name, param, ret)
            else:
                from_final = name
            # node_to
            name, param, ret = self.analyzer.func_prototypes[node_to]
            to_final = format_func_name(name, param, ret)
            if format_fname:
                to_final = format_func_name(name, param, ret)
            else:
                to_final = name
            edges.append(Edge(from_final, to_final, EDGE_CALL))
        log.info('edges: %s', edges)

        return (nodes, edges)

    def visualize_call_flow(self, filename="wasm_call_graph_octopus.gv"):

        nodes, edges = self.get_functions_call_edges(format_fname=False)

        g = Digraph(filename, filename=filename)
        g.attr(rankdir='LR')

        with g.subgraph(name='global') as c:

            export_list = [d['field_str'] for d in self.analyzer.exports]
            import_list = [x for _, x, _ in self.analyzer.imports_func]
            # create all the graph nodes (function name)
            for idx, node in enumerate(nodes):
                if node in import_list:
                    log.debug('import ' + node)
                    fillcolor = DESIGN_IMPORT.get('fillcolor')
                    shape = DESIGN_IMPORT.get('shape')
                    style = DESIGN_IMPORT.get('style')
                    c.node(node, fillcolor=fillcolor, shape=shape, style=style)
                elif node in export_list:
                    log.debug('export ' + node)
                    fillcolor = DESIGN_EXPORT.get('fillcolor')
                    shape = DESIGN_EXPORT.get('shape')
                    style = DESIGN_EXPORT.get('style')
                    c.node(node, fillcolor=fillcolor, shape=shape, style=style)
                else:
                    c.node(node)

            # check if multiple same edges
            # in that case, put the number into label
            edges_counter = dict((x, edges.count(x)) for x in set(edges))
            # insert edges on the graph
            for edge, count in edges_counter.items():
                label = None
                if count > 1:
                    label = str(count)
                c.edge(edge.node_from, edge.node_to, label=label)

        g.render(filename, view=True)



class Function(object):

    def __init__(self, start_offset, start_instr=None,
                 name='func_default_name', prefered_name=None):
        # parameters
        self.start_offset = start_offset
        self.start_instr = start_instr
        self.name = name
        self.prefered_name = prefered_name if prefered_name else name
        self.size = 0
        self.end_offset = None
        self.end_instr = None
        self.basicblocks = list()
        self.instructions = list()

    def show(self):
        line = str(self.start_offset) + ': ' + str(self.name) + '\n'
        line += 'prefered_name: %s' % self.prefered_name
        line += 'start_offset = %x\n' % self.start_offset
        line += 'start_instr = ' + str(self.start_instr.name) + '\n'
        if self.size:
            line += 'size = ' + str(self.size) + '\n'
        if self.end_offset:
            line += 'end_offset = ' + str(self.end_offset) + '\n'
        if self.end_instr:
            line += 'end_instr = ' + str(self.end_instr.name) + '\n'
        line += 'lenght basicblocks: %s' % len(self.basicblocks)
        line += 'lenght instructions: %s' % len(self.instructions)
        line += '\n\n'
        return line
'''
    def show(self):
        attrs = vars(self)
        return '\n'.join("%s: %s" % item for item in attrs.items())
'''



import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError as RequestsConnectionError

import json


class Explorer(object):
    '''
    Generic JSON-RPC client class
    '''

    def __init__(self, host='localhost', port=8332, tls=False, max_retries=3):
        self.host = host
        self.port = port
        self.tls = tls
        self.session = requests.Session()
        self.session.mount(self.host, HTTPAdapter(max_retries=max_retries))
        self.headers = {'Content-Type': 'application/json'}

        self.scheme = 'https' if self.tls else 'http'

        self.url = '{}://{}:{}'.format(self.scheme, self.host, self.port)

    def call(self, method, params=None, jsonrpc=None, _id=None):

        params = params or []
        jsonrpc = jsonrpc or '2.0'
        data = {
            'jsonrpc': jsonrpc,
            'method': method,
            'params': params,
        }
        if _id is not None:
            data['id'] = _id

        try:
            r = self.session.post(self.url, headers=self.headers, data=json.dumps(data))
        except RequestsConnectionError:
            raise Exception('RPC connection Error')
        if not 200 <= r.status_code < 300:
            raise Exception('RPC connection failure: ' + str(r.status_code) + ' ' + r.reason)
        try:
            response = r.json()
        except ValueError:
            raise Exception('JSON response parsing error: ' + str(r.text))
        try:
            return response['result']
        except KeyError:
            raise Exception('\"result\" field in JSON response error: ' + str(response))

    #######################
    # HIGHT-LEVEL METHODS #
    #######################

    def get_transaction(self, transaction_id, verbosity):
        '''
        return transaction informations

        ex :
            binding of eth_getTransactionByHash() for Ethereum
            binding of getrawtransaction() for Bitcoin
            binding of getrawtransaction() for Neo
        '''
        return NotImplementedError

    def get_block_by_number(self, block_number):
        '''
        return block information using given block number

        ex :
            binding of eth_getBlockByNumber() for Ethereum
            binding of getblockhash() + getblock() for Bitcoin
            binding of getblock() for Neo
        '''
        return NotImplementedError

    def get_block_by_hash(self, block_hash):
        '''
        return block information using given block hash

        ex :
            binding of eth_getBlockByHash() for Ethereum
            binding of getblock() for Bitcoin
            binding of getblock() for Neo
        '''
        return NotImplementedError

    def decode_tx(txid):
        '''
        return dict with important transaction information
        '''
        return NotImplementedError
